# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A CLI tool for leveldb files."""
import argparse
import dataclasses
from datetime import datetime
import json
import pathlib

from dfindexeddb import version
from dfindexeddb.leveldb import descriptor
from dfindexeddb.leveldb import ldb
from dfindexeddb.leveldb import log
from dfindexeddb.leveldb import record


_VALID_PRINTABLE_CHARACTERS = (
    ' abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' +
    '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~.')


class Encoder(json.JSONEncoder):
  """A JSON encoder class for dfleveldb fields."""

  def default(self, o):
    """Returns a serializable object for o."""
    if dataclasses.is_dataclass(o):
      o_dict = dataclasses.asdict(o)
      return o_dict
    if isinstance(o, bytes):
      out = []
      for x in o:
        if chr(x) not in _VALID_PRINTABLE_CHARACTERS:
          out.append(f'\\x{x:02X}')
        else:
          out.append(chr(x))
      return ''.join(out)
    if isinstance(o, datetime):
      return o.isoformat()
    if isinstance(o, set):
      return list(o)
    return json.JSONEncoder.default(self, o)


def _Output(structure, output):
  """Helper method to output parsed structure to stdout."""
  if output == 'json':
    print(json.dumps(structure, indent=2, cls=Encoder))
  elif output == 'jsonl':
    print(json.dumps(structure, cls=Encoder))
  elif output == 'repr':
    print(structure)


def DbCommand(args):
  """The CLI for processing leveldb folders."""
  for leveldb_record in record.FolderReader(
      args.source).GetRecords(use_manifest=args.use_manifest):
    _Output(leveldb_record, output=args.output)


def LdbCommand(args):
  """The CLI for processing ldb files."""
  ldb_file = ldb.FileReader(args.source)

  if args.structure_type == 'blocks':
    # Prints block information.
    for block in ldb_file.GetBlocks():
      _Output(block, output=args.output)

  elif args.structure_type == 'records' or not args.structure_type:
    # Prints key value record information.
    for key_value_record in ldb_file.GetKeyValueRecords():
      _Output(key_value_record, output=args.output)

  else:
    print(f'{args.structure_type} is not supported for ldb files.')


def LogCommand(args):
  """The CLI for processing log files."""
  log_file = log.FileReader(args.source)

  if args.structure_type == 'blocks':
    # Prints block information.
    for block in log_file.GetBlocks():
      _Output(block, output=args.output)

  elif args.structure_type == 'physical_records':
    # Prints log file physical record information.
    for log_file_record in log_file.GetPhysicalRecords():
      _Output(log_file_record, output=args.output)

  elif args.structure_type == 'write_batches':
    # Prints log file batch information.
    for batch in log_file.GetWriteBatches():
      _Output(batch, output=args.output)

  elif (args.structure_type in ('parsed_internal_key', 'records')
        or not args.structure_type):
    # Prints key value record information.
    for internal_key_record in log_file.GetParsedInternalKeys():
      _Output(internal_key_record, output=args.output)

  else:
    print(f'{args.structure_type} is not supported for log files.')


def DescriptorCommand(args):
  """The CLI for processing descriptor (MANIFEST) files."""
  manifest_file = descriptor.FileReader(args.source)

  if args.version_history:
    for levels in manifest_file.GetVersions():
      _Output(levels, output=args.output)

  elif args.structure_type == 'blocks':
    # Prints block information.
    for block in manifest_file.GetBlocks():
      _Output(block, output=args.output)

  elif args.structure_type == 'physical_records':
    # Prints log file physical record information.
    for log_file_record in manifest_file.GetPhysicalRecords():
      _Output(log_file_record, output=args.output)

  elif (args.structure_type == 'versionedit'
        or not args.structure_type):
    for version_edit in manifest_file.GetVersionEdits():
      _Output(version_edit, output=args.output)

  else:
    print(f'{args.structure_type} is not supported for descriptor files.')

def App():
  """The CLI app entrypoint for parsing leveldb files."""
  parser = argparse.ArgumentParser(
      prog='dfleveldb',
      description='A cli tool for parsing leveldb files',
      epilog=f'Version {version.GetVersion()}')

  subparsers = parser.add_subparsers()

  parser_db = subparsers.add_parser(
      'db', help='Parse a directory as leveldb.')
  parser_db.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source leveldb directory')
  parser_db.add_argument(
      '--use_manifest',
      action='store_true',
      help='Use manifest file to determine active/deleted records.')
  parser_db.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_db.set_defaults(func=DbCommand)

  parser_log = subparsers.add_parser(
      'log', help='Parse a leveldb log file.')
  parser_log.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source leveldb file')
  parser_log.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_log.add_argument(
      '-t',
      '--structure_type',
      choices=[
          'blocks',
          'physical_records',
          'write_batches',
          'parsed_internal_key'],
      help='Parses the specified structure.  Default is parsed_internal_key.')
  parser_log.set_defaults(func=LogCommand)

  parser_ldb = subparsers.add_parser(
      'ldb', help='Parse a leveldb table (.ldb) file.')
  parser_ldb.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source leveldb file')
  parser_ldb.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_ldb.add_argument(
      '-t',
      '--structure_type',
      choices=[
          'blocks',
          'records'],
      help='Parses the specified structure.  Default is records.')
  parser_ldb.set_defaults(func=LdbCommand)

  parser_descriptor = subparsers.add_parser(
      'descriptor', help='Parse a leveldb descriptor (MANIFEST) file.')
  parser_descriptor.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source leveldb file')
  parser_descriptor.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  db_group = parser_descriptor.add_mutually_exclusive_group()
  db_group.add_argument(
      '-t',
      '--structure_type',
      choices=[
          'blocks', 'physical_records', 'versionedit'],
      help='Parses the specified structure.  Default is versionedit.')
  db_group.add_argument(
      '-v',
      '--version_history',
      action='store_true',
      help='Parses the leveldb version history.')
  parser_descriptor.set_defaults(func=DescriptorCommand)

  args = parser.parse_args()

  if not hasattr(args, 'func'):
    parser.print_usage()
  else:
    args.func(args)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for IndexedDB."""
from enum import Enum, IntEnum, IntFlag


class DatabaseMetaDataKeyType(IntEnum):
  """Database Metadata key types."""
  ORIGIN_NAME = 0
  DATABASE_NAME = 1
  IDB_STRING_VERSION_DATA = 2
  MAX_ALLOCATED_OBJECT_STORE_ID = 3
  IDB_INTEGER_VERSION = 4
  BLOB_NUMBER_GENERATOR_CURRENT_NUMBER = 5
  OBJECT_STORE_META_DATA = 50
  INDEX_META_DATA = 100
  OBJECT_STORE_FREE_LIST = 150
  INDEX_FREE_LIST = 151
  OBJECT_STORE_NAMES = 200
  INDEX_NAMES = 201


class GlobalMetadataKeyType(IntEnum):
  """Global Metadata key types."""
  SCHEMA_VERSION = 0
  MAX_DATABASE_ID = 1
  DATA_VERSION = 2
  RECOVERY_BLOB_JOURNAL = 3
  ACTIVE_BLOB_JOURNAL = 4
  EARLIEST_SWEEP = 5
  EARLIEST_COMPACTION_TIME = 6
  SCOPES_PREFIX = 50
  DATABASE_FREE_LIST = 100
  DATABASE_NAME = 201


class IDBKeyPathType(IntEnum):
  """IndexedDB key path types."""
  NULL = 0
  STRING = 1
  ARRAY = 2


class IDBKeyType(IntEnum):
  """IndexedDB key types."""
  NULL = 0
  STRING = 1
  DATE = 2
  NUMBER = 3
  ARRAY = 4
  MIN_KEY = 5
  BINARY = 6


class IndexMetaDataKeyType(IntEnum):
  """IndexedDB metadata key types."""
  INDEX_NAME = 0
  UNIQUE_FLAG = 1
  KEY_PATH = 2
  MULTI_ENTRY_FLAG = 3


class KeyPrefixType(Enum):
  """IndexedDB key prefix types."""
  GLOBAL_METADATA = 0
  DATABASE_METADATA = 1
  OBJECT_STORE_DATA = 2
  EXISTS_ENTRY = 3
  INDEX_DATA = 4
  INVALID_TYPE = 5
  BLOB_ENTRY = 6


class ObjectStoreMetaDataKeyType(IntEnum):
  """IndexedDB object store metadata key types."""
  OBJECT_STORE_NAME = 0
  KEY_PATH = 1
  AUTO_INCREMENT_FLAG = 2
  IS_EVICTABLE = 3
  LAST_VERSION_NUMBER = 4
  MAXIMUM_ALLOCATED_INDEX_ID = 5
  HAS_KEY_PATH = 6
  KEY_GENERATOR_CURRENT_NUMBER = 7


class ExternalObjectType(IntEnum):
  """IndexedDB external object types."""
  BLOB = 0
  FILE = 1
  FILE_SYSTEM_ACCESS_HANDLE = 2


class BlinkSerializationTag(IntEnum):
  """Blink Javascript serialization tags."""
  MESSAGE_PORT = ord('M')
  MOJO_HANDLE = ord('h')
  BLOB = ord('b')
  BLOB_INDEX = ord('i')
  FILE = ord('f')
  FILE_INDEX = ord('e')
  DOM_FILE_SYSTEM = ord('d')
  FILE_SYSTEM_FILE_HANDLE = ord('n')
  FILE_SYSTEM_DIRECTORY_HANDLE = ord('N')
  FILE_LIST = ord('l')
  FILE_LIST_INDEX = ord('L')
  IMAGE_DATA = ord('#')
  IMAGE_BITMAP = ord('g')
  IMAGE_BITMAP_TRANSFER = ord('G')
  OFFSCREEN_CANVAS_TRANSFER = ord('H')
  READABLE_STREAM_TRANSFER = ord('r')
  TRANSFORM_STREAM_TRANSFER = ord('m')
  WRITABLE_STREAM_TRANSFER = ord('w')
  MEDIA_STREAM_TRACK = ord('s')
  DOM_POINT = ord('Q')
  DOM_POINT_READ_ONLY = ord('W')
  DOM_RECT = ord('E')
  DOM_RECT_READ_ONLY = ord('R')
  DOM_QUAD = ord('T')
  DOM_MATRIX = ord('Y')
  DOM_MATRIX_READ_ONLY = ord('U')
  DOM_MATRIX2D = ord('I')
  DOM_MATRIX2D_READ_ONLY = ord('O')
  CRYPTO_KEY = ord('K')
  RTC_CERTIFICATE = ord('k')
  RTC_ENCODED_AUDIO_FRAME = ord('A')
  RTC_ENCODED_VIDEO_FRAME = ord('V')
  AUDIO_DATA = ord('a')
  VIDEO_FRAME = ord('v')
  ENCODED_AUDIO_CHUNK = ord('y')
  ENCODED_VIDEO_CHUNK = ord('z')
  CROP_TARGET = ord('c')
  RESTRICTION_TARGET = ord('D')
  MEDIA_SOURCE_HANDLE = ord('S')
  DEPRECATED_DETECTED_BARCODE = ord('B')
  DEPRECATED_DETECTED_FACE = ord('F')
  DEPRECATED_DETECTED_TEXT = ord('t')
  FENCED_FRAME_CONFIG = ord('C')
  DOM_EXCEPTION = ord('x')
  TRAILER_OFFSET = 0xFE
  VERSION = 0xFF
  TRAILER_REQUIRES_INTERFACES = 0xA0


class CryptoKeyAlgorithm(IntEnum):
  """CryptoKey Algorithm types."""
  AES_CBC = 1
  HMAC = 2
  RSA_SSA_PKCS1_V1_5 = 3
  SHA1 = 5
  SHA256 = 6
  SHA384 = 7
  SHA512 = 8
  AES_GCM = 9
  RSA_OAEP = 10
  AES_CTR = 11
  AES_KW = 12
  RSA_PSS = 13
  ECDSA = 14
  ECDH = 15
  HKDF = 16
  PBKDF2 = 17
  ED25519 = 18


class NamedCurve(IntEnum):
  """Named Curve types."""
  P256 = 1
  P384 = 2
  P521 = 3


class CryptoKeyUsage(IntFlag):
  """CryptoKey Usage flags."""
  EXTRACTABLE = 1 << 0
  ENCRYPT = 1 << 1
  DECRYPT = 1 << 2
  SIGN = 1 << 3
  VERIFY = 1 << 4
  DERIVE_KEY = 1 << 5
  WRAP_KEY = 1 << 6
  UNWRAP_KEY = 1 << 7
  DRIVE_BITS = 1 << 8


class CryptoKeySubTag(IntEnum):
  """CryptoKey subtag types."""
  AES_KEY = 1
  HMAC_KEY = 2
  RSA_HASHED_KEY = 4
  EC_KEY = 5
  NO_PARAMS_KEY = 6
  ED25519_KEY = 7


class AsymmetricCryptoKeyType(IntEnum):
  """Asymmetric CryptoKey types."""
  PUBLIC_KEY = 1
  PRIVATE_KEY = 2


class WebCryptoKeyType(Enum):
  """WebCryptoKey types."""
  SECRET = 'Secret'
  PUBLIC = 'Public'
  PRIVATE = 'Private'


class V8SerializationTag(IntEnum):
  """V8 Javascript serialization tags."""
  VERSION = 0xFF
  PADDING = ord('\0')
  VERIFY_OBJECT_COUNT = ord('?')
  THE_HOLE = ord('-')
  UNDEFINED = ord('_')
  NULL = ord('0')
  TRUE = ord('T')
  FALSE = ord('F')
  INT32 = ord('I')
  UINT32 = ord('U')
  DOUBLE = ord('N')
  BIGINT = ord('Z')
  UTF8_STRING = ord('S')
  ONE_BYTE_STRING = ord('"')
  TWO_BYTE_STRING = ord('c')
  OBJECT_REFERENCE = ord('^')
  BEGIN_JS_OBJECT = ord('o')
  END_JS_OBJECT = ord('{')
  BEGIN_SPARSE_JS_ARRAY = ord('a')
  END_SPARSE_JS_ARRAY = ord('@')
  BEGIN_DENSE_JS_ARRAY = ord('A')
  END_DENSE_JS_ARRAY = ord('$')
  DATE = ord('D')
  TRUE_OBJECT = ord('y')
  FALSE_OBJECT = ord('x')
  NUMBER_OBJECT = ord('n')
  BIGINT_OBJECT = ord('z')
  STRING_OBJECT = ord('s')
  REGEXP = ord('R')
  BEGIN_JS_MAP = ord(';')
  END_JS_MAP = ord(':')
  BEGIN_JS_SET = ord('\'')
  END_JS_SET = ord(',')
  ARRAY_BUFFER = ord('B')
  RESIZABLE_ARRAY_BUFFER = ord('~')
  ARRAY_BUFFER_TRANSFER = ord('t')
  ARRAY_BUFFER_VIEW = ord('V')
  SHARED_ARRAY_BUFFER = ord('u')
  SHARED_OBJECT = ord('p')
  WASM_MODULE_TRANSFER = ord('w')
  HOST_OBJECT = ord('\\')
  WASM_MEMORY_TRANSFER = ord('m')
  ERROR = ord('r')
  LEGACY_RESERVED_MESSAGE_PORT = ord('M')
  LEGACY_RESERVED_BLOB = ord('b')
  LEGACY_RESERVED_BLOB_INDEX = ord('i')
  LEGACY_RESERVED_FILE = ord('f')
  LEGACY_RESERVED_FILE_INDEX = ord('e')
  LEGACY_RESERVED_DOM_FILE_SYSTEM = ord('d')
  LEGACY_RESERVED_FILE_LIST = ord('l')
  LEGACY_RESERVED_FILE_LIST_INDEX = ord('L')
  LEGACY_RESERVED_IMAGE_DATA = ord('#')
  LEGACY_RESERVED_IMAGE_BITMAP = ord('g')
  LEGACY_RESERVED_IMAGE_BITMAP_TRANSFER = ord('G')
  LEGACY_RESERVED_OFF_SCREEN_CANVAS = ord('H')
  LEGACY_RESERVED_CRYPTO_KEY = ord('K')
  LEGACY_RESERVED_RTC_CERTIFICATE = ord('k')


class V8ArrayBufferViewTag(IntEnum):
  """V8 ArrayBufferView tags."""
  INT8_ARRAY = ord('b')
  UINT8_ARRAY = ord('B')
  UINT8_CLAMPED_ARRAY = ord('C')
  INT16_ARRAY = ord('w')
  UINT16_ARRAY = ord('W')
  INT32_ARRAY = ord('d')
  UINT32_ARRAY = ord('D')
  FLOAT32_ARRAY = ord('f')
  FLOAT64_ARRAY = ord('F')
  BIGINT64_ARRAY = ord('q')
  BIGUINT64_ARRAY = ord('Q')
  DATAVIEW = ord('?')


class V8ErrorTag(IntEnum):
  """V8 Error tags."""
  EVAL_ERROR_PROTOTYPE = ord('E')
  RANGE_ERROR_PROTOTYPE = ord('R')
  REFERENCE_ERROR_PROTOTYPE = ord('F')
  SYNTAX_ERROR_PROTOTYPE = ord('S')
  TYPE_ERROR_PROTOTYPE = ord('T')
  URI_ERROR_PROTOTYPE = ord('U')
  MESSAGE = ord('m')
  CAUSE = ord('c')
  STACK = ord('s')
  END = ord('.')


class ImageSerializationTag(IntEnum):
  """Image Serialization tags."""
  END = 0
  PREDEFINED_COLOR_SPACE = 1
  CANVAS_PIXEL_FORMAT = 2
  IMAGE_DATA_STORAGE_FORMAT = 3
  ORIGIN_CLEAN = 4
  IS_PREMULTIPLIED = 5
  CANVAS_OPACITY_MODE = 6
  PARAMETRIC_COLOR_SPACE = 7
  IMAGE_ORIENTATION = 8
  LAST = IMAGE_ORIENTATION


class SerializedPredefinedColorSpace(IntEnum):
  """Serialized Predefined Color Space enumeration."""
  LEGACY_OBSOLETE = 0
  SRGB = 1
  REC2020 = 2
  P3 = 3
  REC2100HLG = 4
  REC2100PQ = 5
  SRGB_LINEAR = 6
  LAST = SRGB_LINEAR


class SerializedPixelFormat(IntEnum):
  """Serialized Pixel Format enumeration."""
  NATIVE8_LEGACY_OBSOLETE = 0
  F16 = 1
  RGBA8 = 2
  BGRA8 = 3
  RGBX8 = 4
  LAST = RGBX8


class SerializedImageDataStorageFormat(IntEnum):
  """The Serialized Image Data Storage Format."""
  UINT8CLAMPED = 0
  UINT16 = 1
  FLOAT32 = 2
  LAST = FLOAT32


class SerializedOpacityMode(IntEnum):
  """The Serialized Opacity Mode."""
  KNONOPAQUE = 0
  KOPAQUE = 1
  KLAST = KOPAQUE


class SerializedImageOrientation(IntEnum):
  """The Serialized Image Orientation."""
  TOP_LEFT = 0
  TOP_RIGHT = 1
  BOTTOM_RIGHT = 2
  BOTTOM_LEFT = 3
  LEFT_TOP = 4
  RIGHT_TOP = 5
  RIGHT_BOTTOM = 6
  LEFT_BOTTOM = 7
  LAST = LEFT_BOTTOM



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for v8 javascript serialized objects."""
from dataclasses import dataclass
from datetime import datetime
import io
import os
from typing import Any, BinaryIO, Dict, Optional, Set, Tuple, Union

from dfindexeddb import errors
from dfindexeddb import utils
from dfindexeddb.indexeddb.chromium import definitions


@dataclass
class BufferArrayView:
  """A parsed Javascript BufferArrayView."""
  buffer: bytes
  tag: definitions.V8ArrayBufferViewTag
  offset: int
  length: int
  flags: int


class JSArray(list):
  """A parsed Javascript array.

  This is a wrapper around a standard Python list to allow assigning arbitrary
  properties as is possible in the Javascript equivalent.
  """

  def __repr__(self):
    array_entries = ", ".join([str(entry) for entry in list(self)])
    properties = ", ".join(
        f'{key}: {value}' for key, value in self.properties.items())
    return f'[{array_entries}, {properties}]'

  @property
  def properties(self) -> Dict[str, Any]:
    """Returns the object properties."""
    return self.__dict__

  def __contains__(self, item):
    return item in self.__dict__

  def __getitem__(self, name):
    return self.__dict__[name]


@dataclass
class Null:
  """A parsed Javascript Null."""


@dataclass
class RegExp:
  """A parsed Javascript RegExp."""
  pattern: str
  flags: int


@dataclass(frozen=True)
class Undefined:
  """A parsed Javascript undef."""


class ValueDeserializer:
  """A class to deserialize v8 Javascript values/objects.

  Attributes:
    decoder (utils.StreamDecoder): A StreamDecoder that wraps the binary
      stream containing the serialized Javascript value/object.
    delegate: An instance to delegate deserializing an object
    next_id (int): the next ID to use for a deserialized object.
    objects (dict[int, Any]): a dictionary mapping integer IDs to a
      deserialized object.
    version (int): the version of the serialized Javascript value/object
  """

  LATEST_VERSION = 15

  def __init__(self, stream: BinaryIO, delegate):
    """Initializes a ValueDeserializer.

    Args:
      stream: a stream of bytes to be deserialized.
      delegate: an object to delegate additional deserialization of additional
        Javascript types/objects.
    """
    self.decoder = utils.StreamDecoder(stream)
    self.delegate = delegate
    self.next_id = 0
    self.objects = {}
    self.version = None

  def GetWireFormatVersion(self) -> int:
    """Returns the underlying wire format version.

    Note: Return value is only valid after a call to ReadHeader.
    """
    return self.version

  def ReadHeader(self) -> bool:
    """Reads the header and returns True if the version is supported."""
    if self._ReadTag() != definitions.V8SerializationTag.VERSION:
      return False
    _, self.version = self.decoder.DecodeUint32Varint()
    if self.version > self.LATEST_VERSION:
      return False
    return True

  def ReadObjectWrapper(self):
    """Deserializes a V8 object from the current decoder position."""
    original_position = self.decoder.stream.tell()
    result = self._ReadObject()
    if result is None and self.version == 13:
      self.decoder.stream.seek(original_position, os.SEEK_SET)
      result = self._ReadObject()
      raise NotImplementedError('ValueDeserializer.ReadObjectWrapper v13')
    return result

  def ReadObjectUsingEntireBufferForLegacyFormat(self):
    """Reads an object, consuming the entire buffer."""
    raise NotImplementedError(
        'ValueDeserializer.ReadObjectUsingEntireBufferForLegacyFormat')

  def ReadValue(self):
    """Reads the Javascript value."""
    if self.GetWireFormatVersion() > 0:
      return self.ReadObjectWrapper()
    return self.ReadObjectUsingEntireBufferForLegacyFormat()

  def _PeekTag(self) -> Optional[definitions.V8SerializationTag]:
    """Peeks the next serialization tag from the current decoder position.

    Returns:
      The serialization tag or None if there is no more bytes to read.
    """
    try:
      _, tag_value = self.decoder.PeekBytes(1)
    except errors.DecoderError:
      return None
    try:
      return definitions.V8SerializationTag(tag_value[0])
    except ValueError as error:
      raise errors.ParserError(
          f'Invalid v8 tag value {tag_value} at offset'
          f' {self.decoder.stream.tell()}') from error

  def _ReadTag(self) -> definitions.V8SerializationTag:
    """Returns the next non-padding serialization tag.

    Raises:
      errors.ParserError: when an invalid v8 tag is read.
    """
    while True:
      _, tag_value = self.decoder.ReadBytes(1)
      try:
        tag = definitions.V8SerializationTag(tag_value[0])
      except ValueError as error:
        raise errors.ParserError(f'Invalid v8 tag value {tag_value}') from error
      if tag != definitions.V8SerializationTag.PADDING:
        return tag

  def _ConsumeTag(self, peeked_tag: definitions.V8SerializationTag):
    """Consumes the next serialization tag.

    Args:
      peeked_tag: the tag that is expected to be read from the current position.

    Raises:
      errors.ParserError: if the peeked tag is not read.
    """
    tag = self._ReadTag()
    if tag != peeked_tag:
      raise errors.ParserError(f'Unexpected tag {tag} found.')

  def _ReadObject(self):
    """Reads a Javascript object from the current position."""
    _, result = self._ReadObjectInternal()
    tag = self._PeekTag()
    if tag and tag == definitions.V8SerializationTag.ARRAY_BUFFER_VIEW:
      self._ConsumeTag(tag)
      result = self._ReadJSArrayBufferView(result)
    return result

  def _ReadObjectInternal(self) -> Tuple[definitions.V8SerializationTag, Any]:
    """Reads a Javascript object from the current position.

    Returns:
      a tuple of the serialization tag and the parsed object.
    """
    tag = self._ReadTag()
    if tag == definitions.V8SerializationTag.VERIFY_OBJECT_COUNT:
      _ = self.decoder.DecodeUint32Varint()
      parsed_object = self._ReadObject()
    elif tag == definitions.V8SerializationTag.UNDEFINED:
      parsed_object = Undefined()
    elif tag == definitions.V8SerializationTag.NULL:
      parsed_object = Null()
    elif tag == definitions.V8SerializationTag.TRUE:
      parsed_object = True
    elif tag == definitions.V8SerializationTag.FALSE:
      parsed_object = False
    elif tag == definitions.V8SerializationTag.INT32:
      _, parsed_object = self.decoder.DecodeInt32Varint()
    elif tag == definitions.V8SerializationTag.UINT32:
      _, parsed_object = self.decoder.DecodeUint32Varint()
    elif tag == definitions.V8SerializationTag.DOUBLE:
      _, parsed_object = self.decoder.DecodeDouble()
    elif tag == definitions.V8SerializationTag.BIGINT:
      parsed_object = self.ReadBigInt()
    elif tag == definitions.V8SerializationTag.UTF8_STRING:
      parsed_object = self.ReadUTF8String()
    elif tag == definitions.V8SerializationTag.ONE_BYTE_STRING:
      parsed_object = self.ReadOneByteString()
    elif tag == definitions.V8SerializationTag.TWO_BYTE_STRING:
      parsed_object = self.ReadTwoByteString()
    elif tag == definitions.V8SerializationTag.OBJECT_REFERENCE:
      _, object_id = self.decoder.DecodeUint32Varint()
      parsed_object = self.objects[object_id]
    elif tag == definitions.V8SerializationTag.BEGIN_JS_OBJECT:
      parsed_object = self._ReadJSObject()
    elif tag == definitions.V8SerializationTag.BEGIN_SPARSE_JS_ARRAY:
      parsed_object = self.ReadSparseJSArray()
    elif tag == definitions.V8SerializationTag.BEGIN_DENSE_JS_ARRAY:
      parsed_object = self.ReadDenseJSArray()
    elif tag == definitions.V8SerializationTag.DATE:
      parsed_object = self._ReadJSDate()
    elif tag in (
        definitions.V8SerializationTag.TRUE_OBJECT,
        definitions.V8SerializationTag.FALSE_OBJECT,
        definitions.V8SerializationTag.NUMBER_OBJECT,
        definitions.V8SerializationTag.BIGINT_OBJECT,
        definitions.V8SerializationTag.STRING_OBJECT):
      parsed_object = self._ReadJSPrimitiveWrapper(tag)
    elif tag == definitions.V8SerializationTag.REGEXP:
      parsed_object = self._ReadJSRegExp()
    elif tag == definitions.V8SerializationTag.BEGIN_JS_MAP:
      parsed_object = self._ReadJSMap()
    elif tag == definitions.V8SerializationTag.BEGIN_JS_SET:
      parsed_object = self._ReadJSSet()
    elif tag == definitions.V8SerializationTag.ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=False, is_resizable=False)
    elif tag == definitions.V8SerializationTag.RESIZABLE_ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=False, is_resizable=True)
    elif tag == definitions.V8SerializationTag.SHARED_ARRAY_BUFFER:
      parsed_object = self._ReadJSArrayBuffer(
          is_shared=True, is_resizable=False)
    elif tag == definitions.V8SerializationTag.ERROR:
      parsed_object = self._ReadJSError()
    elif tag == definitions.V8SerializationTag.WASM_MODULE_TRANSFER:
      parsed_object = self._ReadWasmModuleTransfer()
    elif tag == definitions.V8SerializationTag.WASM_MEMORY_TRANSFER:
      parsed_object = self._ReadWasmMemory()
    elif tag == definitions.V8SerializationTag.HOST_OBJECT:
      parsed_object = self.ReadHostObject()
    elif (
        tag == definitions.V8SerializationTag.SHARED_OBJECT and
            self.version >= 15):
      parsed_object = self.ReadSharedObject()
    elif self.version < 13:
      self.decoder.stream.seek(-1, os.SEEK_CUR)
      parsed_object = self.ReadHostObject()
    else:
      parsed_object = None
    return tag, parsed_object

  def ReadString(self) -> str:
    """Reads a string from the current position.

    Raises:
      errors.ParserError: if the parsed object is not a string.
    """
    if self.version < 12:
      return self.ReadUTF8String()

    str_obj = self._ReadObject()
    if not isinstance(str_obj, str):
      raise errors.ParserError('Not a string')
    return str_obj

  def ReadBigInt(self) -> int:
    """Reads a Javascript Bigint from the current position."""
    bit_field = self.decoder.DecodeUint32Varint()[1]
    byte_count = bit_field >> 1
    signed = bool(bit_field & 0x1)
    _, bigint = self.decoder.DecodeInt(byte_count=byte_count)
    return -bigint if signed else bigint

  def ReadUTF8String(self) -> str:
    """Reads a UTF-8 string from the current position."""
    count = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=count)[1]
    return buffer.decode('utf-8')

  def ReadOneByteString(self) -> str:
    """Reads a one-byte string from the current position.

    The raw bytes are decoded using latin-1 encoding.
    """
    length = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=length)[1]
    return buffer.decode('latin-1')

  def ReadTwoByteString(self) -> str:
    """Reads a UTF-16-LE string from the current position."""
    length = self.decoder.DecodeUint32Varint()[1]
    buffer = self.decoder.ReadBytes(count=length)[1]
    return buffer.decode('utf-16-le')

  def ReadExpectedString(self) -> Optional[str]:
    """Reads a string from the current position, None if there is no tag or
      invalid byte length.

    A string can be UTF-8/one-byte/two-byte.

    Raises:
      errors.ParserError: if there is an unexpected serialization tag."""
    tag = self._ReadTag()
    if not tag:
      return None
    _, byte_length = self.decoder.DecodeUint32Varint()
    if not byte_length:
      return None

    _, raw_bytes = self.decoder.ReadBytes(count=byte_length)

    if tag == definitions.V8SerializationTag.ONE_BYTE_STRING:
      return raw_bytes.decode('latin-1')
    if tag == definitions.V8SerializationTag.TWO_BYTE_STRING:
      return raw_bytes.decode('utf-16-le')
    if tag == definitions.V8SerializationTag.UTF8_STRING:
      return raw_bytes.decode('utf-8')
    raise errors.ParserError(f'Unexpected serialization tag {tag}.')

  def _ReadJSObject(self) -> Dict:
    """Reads a JSObject from the current position.

    Raises:
      errors.ParserError: if an unexpected number of properties have been
        read.
    """
    next_id = self._GetNextId()
    js_object = {}

    num_properties = self._ReadJSObjectProperties(
        js_object, definitions.V8SerializationTag.END_JS_OBJECT)
    _, expected_number_properties = self.decoder.DecodeUint32Varint()
    if expected_number_properties != num_properties:
      raise errors.ParserError('Unexpected number of properties')

    self.objects[next_id] = js_object
    return js_object

  def _ReadJSObjectProperties(
      self,
      js_object: Union[Dict, JSArray],
      end_tag: definitions.V8SerializationTag
  ) -> int:
    """Reads key-value properties and sets them to the given js_object.

    Args:
      js_object: the Javascript object to set the parsed properties to.
      end_tag: the end tag that signifies there are no more properties to read.

    Returns:
      the number of properties that were read.
    """
    num_properties = 0
    while self._PeekTag() != end_tag:
      key = self._ReadObject()
      value = self._ReadObject()
      js_object[key] = value
      num_properties += 1
    self._ConsumeTag(end_tag)
    return num_properties

  def _GetNextId(self) -> int:
    """Gets the next object ID."""
    next_id = self.next_id
    self.next_id += 1
    return next_id

  def ReadSparseJSArray(self) -> JSArray:
    """Reads a sparse encoded JSArray from the current position.

    Raises:
      errors.ParserError: if there is an unexpected property or array length.
    """
    next_id = self._GetNextId()

    js_array = JSArray()
    _, length = self.decoder.DecodeUint32Varint()
    for _ in range(length):
      js_array.append(Undefined())

    num_properties = self._ReadJSObjectProperties(
        js_array.__dict__, definitions.V8SerializationTag.END_SPARSE_JS_ARRAY)
    _, expected_num_properties = self.decoder.DecodeUint32Varint()
    _, expected_length = self.decoder.DecodeUint32Varint()

    if num_properties != expected_num_properties:
      raise errors.ParserError('Unexpected property length')
    if length != expected_length:
      raise errors.ParserError('Unexpected array length')
    self.objects[next_id] = js_array
    return js_array

  def ReadDenseJSArray(self) -> JSArray:
    """Reads a dense encoded JSArray from the current position.

    Raises:
      errors.ParserError: if there is an unexpected property or array length.
    """
    next_id = self._GetNextId()

    js_array = JSArray()
    _, length = self.decoder.DecodeUint32Varint()
    for _ in range(length):
      tag = self._PeekTag()
      if tag == definitions.V8SerializationTag.THE_HOLE:
        self._ConsumeTag(tag)
        continue
      array_object = self._ReadObject()

      if self.version < 11 and isinstance(array_object, Undefined):
        continue
      js_array.append(array_object)

    num_properties = self._ReadJSObjectProperties(
        js_array.__dict__, definitions.V8SerializationTag.END_DENSE_JS_ARRAY)
    _, expected_num_properties = self.decoder.DecodeUint32Varint()
    _, expected_length = self.decoder.DecodeUint32Varint()
    if num_properties != expected_num_properties:
      raise errors.ParserError('Unexpected property length')
    if length != expected_length:
      raise errors.ParserError('Unexpected array length')
    self.objects[next_id] = js_array
    return js_array

  def _ReadJSDate(self) -> datetime:
    """Reads a JSDate from the current position."""
    next_id = self._GetNextId()

    _, value = self.decoder.DecodeDouble()
    result = datetime.utcfromtimestamp(value/1000.0)
    self.objects[next_id] = result
    return result

  def _ReadJSPrimitiveWrapper(
      self,
      tag: definitions.V8SerializationTag
  ) -> Union[bool, float, int, str]:
    """Reads a Javascript wrapped primitive.

    Args:
      tag: the serialization tag.

    Returns:
      the parsed value.

    Raises:
      errors.ParserError: if the given tag is not a Javascript wrapped
        primitive.
    """
    next_id = self._GetNextId()

    if tag == definitions.V8SerializationTag.TRUE_OBJECT:
      value = True
    elif tag == definitions.V8SerializationTag.FALSE_OBJECT:
      value = False
    elif tag == definitions.V8SerializationTag.NUMBER_OBJECT:
      _, value = self.decoder.DecodeDouble()
    elif tag == definitions.V8SerializationTag.BIGINT_OBJECT:
      value = self.ReadBigInt()
    elif tag == definitions.V8SerializationTag.STRING_OBJECT:
      value = self.ReadString()
    else:
      raise errors.ParserError(f'Invalid tag {tag}')

    self.objects[next_id] = value
    return value

  def _ReadJSRegExp(self) -> RegExp:
    """Reads a Javascript regular expression from the current position."""
    next_id = self._GetNextId()
    pattern = self.ReadString()
    _, flags = self.decoder.DecodeUint32Varint()  # TODO: verify flags
    regexp = RegExp(pattern=pattern, flags=flags)
    self.objects[next_id] = regexp
    return regexp

  def _ReadJSMap(self) -> Dict[Any, Any]:
    """Reads a Javascript map (dictionary) from the current position.

    Raises:
      errors.ParserError: on an unexpected map length.
    """
    next_id = self._GetNextId()
    js_map = {}

    tag = self._PeekTag()
    while tag != definitions.V8SerializationTag.END_JS_MAP:
      key = self._ReadObject()
      value = self._ReadObject()
      js_map[key] = value
      tag = self._PeekTag()
    self._ConsumeTag(definitions.V8SerializationTag.END_JS_MAP)

    _, expected_length = self.decoder.DecodeUint32Varint()
    if len(js_map) * 2 != expected_length:
      raise errors.ParserError('unexpected length')

    self.objects[next_id] = js_map
    return js_map

  def _ReadJSSet(self) -> Set[Any]:
    """Reads a Javascript set from the current position.

    Raises:
      errors.ParserError: on an unexpected map length.
    """
    next_id = self._GetNextId()

    js_set = set()
    tag = self._PeekTag()
    while tag != definitions.V8SerializationTag.END_JS_SET:
      element = self._ReadObject()
      js_set.add(element)
      tag = self._PeekTag()
    self._ConsumeTag(definitions.V8SerializationTag.END_JS_SET)

    _, expected_length = self.decoder.DecodeUint32Varint()
    if len(js_set) != expected_length:
      raise ValueError('unexpected length')

    self.objects[next_id] = js_set
    return js_set

  def _ReadJSArrayBuffer(
      self, is_shared: bool, is_resizable: bool) -> bytes:
    """Reads a Javascript ArrayBuffer from the current position.

    Args:
      is_shared: True if the buffer is shared, False otherwise.
      is_resizable: True if the buffer is resizable, False otherwise.
    """
    next_id = self._GetNextId()
    array_buffer = b''

    if is_shared:
      raise NotImplementedError('Shared ArrayBuffer not supported yet')

    _, byte_length = self.decoder.DecodeUint32Varint()
    max_byte_length = byte_length
    if is_resizable:
      _, max_byte_length = self.decoder.DecodeUint32Varint()
      if byte_length > max_byte_length:
        return array_buffer
    if byte_length:
      _, array_buffer = self.decoder.ReadBytes(byte_length)

    self.objects[next_id] = array_buffer
    return array_buffer

  def _ReadJSArrayBufferView(self, buffer):
    """Reads a JSArrayBufferView from the current position."""
    _, tag = self.decoder.ReadBytes(1)
    _, byte_offset = self.decoder.DecodeUint32Varint()
    _, byte_length = self.decoder.DecodeUint32Varint()

    if self.version >= 14:  # or version_13_broken_data_mode
      _, flags = self.decoder.DecodeUint32Varint()
    else:
      flags = 0

    return BufferArrayView(
        buffer=buffer,
        tag=definitions.V8ArrayBufferViewTag(tag[0]),
        offset=byte_offset,
        length=byte_length,
        flags=flags)

  def _ReadJSError(self):
    """Reads a Javascript error from the current position."""
    raise NotImplementedError('ValueDeserializer.ReadJSError')

  def _ReadWasmModuleTransfer(self):
    """Reads a Wasm module transfer object from the current position."""
    raise NotImplementedError('ValueDeserializer.ReadWasmModuleTransfer')

  def _ReadWasmMemory(self):
    """Reads a Wasm memory object from the current position."""
    raise NotImplementedError('ValueDeserializer.ReadWasmMemory')

  def ReadSharedObject(self) -> Any:
    """Reads a shared object from the current position."""
    #_, shared_object_id = self.decoder.DecodeUint32Varint()
    raise NotImplementedError('ValueDeserializer.ReadSharedObject')

  def ReadHostObject(self):
    """Reads a Host object using the delegate object.

    Raises:
      errors.ParserError: when the delegate object is None.
    """
    next_id = self._GetNextId()
    if not self.delegate:
      raise errors.ParserError('No delegate to read host object.')
    host_object = self.delegate.ReadHostObject()
    self.objects[next_id] = host_object
    return host_object

  @classmethod
  def FromBytes(cls, data: bytes, delegate: Any) -> Any:
    """Returns a deserialized javascript object from the data.

    Args:
      data: the data to deserialize/parse.
      delegate: an object to delegate parsing additional Javascript objects.

    Returns:
      A python representation of the parsed javascript object.

    Raises:
      errors.ParserError: if there is an invalid V8 javascript header.
    """
    stream = io.BytesIO(data)
    deserializer = cls(stream, delegate)
    if not deserializer.ReadHeader():
      raise errors.ParserError('Invalid V8 header')
    return deserializer.ReadObjectWrapper()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A CLI tool for dfindexeddb."""
import argparse
import dataclasses
import enum
from datetime import datetime
import json
import pathlib

from dfindexeddb import version
from dfindexeddb.indexeddb.chromium import blink
from dfindexeddb.indexeddb.chromium import record as chromium_record
from dfindexeddb.indexeddb.chromium import v8
from dfindexeddb.indexeddb.safari import record as safari_record


_VALID_PRINTABLE_CHARACTERS = (
    ' abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' +
    '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~.')


class Encoder(json.JSONEncoder):
  """A JSON encoder class for dfindexeddb fields."""
  def default(self, o):
    if dataclasses.is_dataclass(o):
      o_dict = dataclasses.asdict(o)
      return o_dict
    if isinstance(o, bytes):
      out = []
      for x in o:
        if chr(x) not in _VALID_PRINTABLE_CHARACTERS:
          out.append(f'\\x{x:02X}')
        else:
          out.append(chr(x))
      return ''.join(out)
    if isinstance(o, datetime):
      return o.isoformat()
    if isinstance(o, v8.Undefined):
      return "<undefined>"
    if isinstance(o, v8.Null):
      return "<null>"
    if isinstance(o, set):
      return list(o)
    if isinstance(o, v8.RegExp):
      return str(o)
    if isinstance(o, enum.Enum):
      return o.name
    return json.JSONEncoder.default(self, o)


def _Output(structure, output):
  """Helper method to output parsed structure to stdout."""
  if output == 'json':
    print(json.dumps(structure, indent=2, cls=Encoder))
  elif output == 'jsonl':
    print(json.dumps(structure, cls=Encoder))
  elif output == 'repr':
    print(structure)


def BlinkCommand(args):
  """The CLI for processing a file as a blink value."""
  with open(args.source, 'rb') as fd:
    buffer = fd.read()
    blink_value = blink.V8ScriptValueDecoder.FromBytes(buffer)
    _Output(blink_value, output=args.output)


def DbCommand(args):
  """The CLI for processing a directory as IndexedDB."""
  if args.format in ('chrome', 'chromium'):
    for db_record in chromium_record.FolderReader(
        args.source).GetRecords(use_manifest=args.use_manifest):
      _Output(db_record, output=args.output)
  elif args.format == 'safari':
    for db_record in safari_record.FileReader(args.source).Records():
      _Output(db_record, output=args.output)


def LdbCommand(args):
  """The CLI for processing a LevelDB table (.ldb) file as IndexedDB."""
  for db_record in chromium_record.IndexedDBRecord.FromFile(args.source):
    _Output(db_record, output=args.output)


def LogCommand(args):
  """The CLI for processing a LevelDB log file as IndexedDB."""
  for db_record in chromium_record.IndexedDBRecord.FromFile(args.source):
    _Output(db_record, output=args.output)


def App():
  """The CLI app entrypoint for dfindexeddb."""
  parser = argparse.ArgumentParser(
      prog='dfindexeddb',
      description='A cli tool for parsing IndexedDB files',
      epilog=f'Version {version.GetVersion()}')

  subparsers = parser.add_subparsers()

  parser_blink = subparsers.add_parser(
      'blink', help='Parse a file as a blink value.')
  parser_blink.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help=(
        'The source file.'))
  parser_blink.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_blink.set_defaults(func=BlinkCommand)

  parser_db = subparsers.add_parser(
      'db', help='Parse a directory as IndexedDB.')
  parser_db.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help=(
        'The source IndexedDB folder (for chrome/chromium) '
        'or file (for safari).'))
  parser_db.add_argument(
      '--format',
      required=True,
      choices=['chromium', 'chrome', 'safari'],
      help='The type of IndexedDB to parse.')
  parser_db.add_argument(
      '--use_manifest',
      action='store_true',
      help='Use manifest file to determine active/deleted records.')
  parser_db.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_db.set_defaults(func=DbCommand)

  parser_ldb = subparsers.add_parser(
      'ldb',
      help='Parse a ldb file as IndexedDB.')
  parser_ldb.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source .ldb file.')
  parser_ldb.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_ldb.set_defaults(func=LdbCommand)

  parser_log = subparsers.add_parser(
      'log',
      help='Parse a log file as IndexedDB.')
  parser_log.add_argument(
      '-s', '--source',
      required=True,
      type=pathlib.Path,
      help='The source .log file.')
  parser_log.add_argument(
      '-o',
      '--output',
      choices=[
          'json',
          'jsonl',
          'repr'],
      default='json',
      help='Output format.  Default is json')
  parser_log.set_defaults(func=LogCommand)

  args = parser.parse_args()
  if hasattr(args, 'func'):
    args.func(args)
  else:
    parser.print_help()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for WebKit encoded JavaScript values."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import io
import plistlib
from typing import Any, Dict, List, Tuple, Union

from dfindexeddb import errors
from dfindexeddb import utils
from dfindexeddb.indexeddb.safari import definitions


@dataclass
class ArrayBufferView:
  """A parsed JavaScript ArrayBufferView.

  Attributes:
    array_buffer_view_subtag: the sub tag.
    buffer: the buffer.
    offset: the offset of the view.
    length: the length of the view.
  """
  array_buffer_view_subtag: definitions.ArrayBufferViewSubtag
  buffer: bytes
  offset: int
  length: int


@dataclass
class ResizableArrayBuffer:
  """A parsed Resizable Array Buffer.

  Attributes:
    buffer: the buffer.
    max_length: the maximum length of the buffer (for resizing).
  """
  buffer: bytes
  max_length: int


@dataclass
class FileData:
  """A parsed FileData.

  Attributes:
    path: the path.
    url: the URL.
    type: the type.
    name: the file name.
    last_modified: the last modified timestamp.
  """
  path: str
  url: str
  type: str
  name: str
  last_modified: float


@dataclass
class FileList:
  """A parsed FileList.

  Attributes:
    files: the list of files.
  """
  files: List[FileData]


class JSArray(list):
  """A parsed JavaScript array.

  This is a wrapper around a standard Python list to allow assigning arbitrary
  properties as is possible in the JavaScript equivalent.
  """

  def __repr__(self):
    array_entries = ", ".join([str(entry) for entry in list(self)])
    properties = ", ".join(
        f'{key}: {value}' for key, value in self.properties.items())
    return f'[{array_entries}, {properties}]'

  @property
  def properties(self) -> Dict[str, Any]:
    """Returns the object properties."""
    return self.__dict__

  def __contains__(self, item):
    return item in self.__dict__

  def __getitem__(self, name):
    return self.__dict__[name]


class JSSet(set):
  """A parsed JavaScript set.

  This is a wrapper around a standard Python set to allow assigning arbitrary
  properties as is possible in the JavaScript equivalent.
  """

  def __repr__(self):
    set_entries = ", ".join([str(entry) for entry in list(self)])
    properties = ", ".join(
        f'{key}: {value}' for key, value in self.properties.items())
    return f'[{set_entries}, {properties}]'

  @property
  def properties(self) -> Dict[str, Any]:
    """Returns the object properties."""
    return self.__dict__

  def __contains__(self, item):
    return item in self.__dict__

  def __getitem__(self, name):
    return self.__dict__[name]


@dataclass
class Null:
  """A parsed JavaScript Null."""


@dataclass
class RegExp:
  """A parsed JavaScript RegExp.

  Attributes:
    pattern: the pattern.
    flags: the flags.
  """
  pattern: str
  flags: str


@dataclass(frozen=True)
class Undefined:
  """A parsed JavaScript undef."""


@dataclass
class IDBKeyData(utils.FromDecoderMixin):
  """An IDBKeyData.

  Attributes:
    offset: the offset at which the IDBKeyData was parsed.
    key_type: the IDB Key Type.
    data: the key data.
  """
  offset: int
  key_type: definitions.SIDBKeyType
  data: Union[float, datetime, str, bytes, list]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.StreamDecoder, base_offset: int = 0) -> IDBKeyData:
    """Decodes an IDBKeyData from the current position of decoder.

    Refer to IDBSerialization.cpp for the encoding scheme.

    Args:
      decoder: the decoder

    Returns:
      the IDBKeyData.

    Raises:
      ParserError: when the key version is not found or an unknown key type is
          encountered or an old-style PropertyList key type is found.
    """
    def _DecodeKeyBuffer(key_type):
      if key_type == definitions.SIDBKeyType.MIN:
        data = None
      if key_type == definitions.SIDBKeyType.NUMBER:
        _, data = decoder.DecodeDouble()
      elif key_type == definitions.SIDBKeyType.DATE:
        _, timestamp = decoder.DecodeDouble()
        data = datetime.utcfromtimestamp(timestamp/1000)
      elif key_type == definitions.SIDBKeyType.STRING:
        _, length = decoder.DecodeUint32()
        _, raw_data = decoder.ReadBytes(length*2)
        data = raw_data.decode('utf-16-le')
      elif key_type == definitions.SIDBKeyType.BINARY:
        _, length = decoder.DecodeUint32()
        _, data = decoder.ReadBytes(length)
      elif key_type == definitions.SIDBKeyType.ARRAY:
        _, length = decoder.DecodeUint64()
        data = []
        for _ in range(length):
          _, key_type = decoder.DecodeUint8()
          element = _DecodeKeyBuffer(key_type)
          data.append(element)
      else:
        raise errors.ParserError('Unknown definitions.SIDBKeyType found.')
      return data

    offset, version_header = decoder.DecodeUint8()
    if version_header != definitions.SIDBKeyVersion:
      raise errors.ParserError('SIDBKeyVersion not found.')

    _, raw_key_type = decoder.DecodeUint8()
    key_type = definitions.SIDBKeyType(raw_key_type)

    # "Old-style key is characterized by this magic character that
    # begins serialized PropertyLists
    if key_type == b'b':
      raise errors.ParserError('Old-style PropertyList key type found.')
    data = _DecodeKeyBuffer(key_type)

    return cls(
        offset=offset+base_offset,
        key_type=key_type,
        data=data)


class SerializedScriptValueDecoder():
  """Decodes a Serialized Script Value from a stream of bytes.

  Attributes:
    decoder: the stream decoder for the given byte stream.
    version: the parsed serialized script version.
    constant_pool: the constant pool.
    object_pool: the object pool.
  """
  def __init__(self, stream: io.BytesIO):
    self.decoder = utils.StreamDecoder(stream)
    self.version = None
    self.constant_pool = []
    self.object_pool = []

  def PeekTag(self) -> int:
    """Peeks a tag from the current position."""
    _, peeked_bytes = self.decoder.PeekBytes(4)
    return int.from_bytes(peeked_bytes, byteorder='little')

  def PeekSerializationTag(self) -> definitions.SerializationTag:
    """Peeks a SerializationTag from the current position.

    Raises:
      ParserError if an invalid SerializationTag was parsed.
    """
    offset, terminal_byte = self.decoder.PeekBytes(1)
    try:
      return definitions.SerializationTag(terminal_byte[0])
    except ValueError as error:
      raise errors.ParserError(
          f'Invalid SerializationTag {terminal_byte} at offset {offset}'
      ) from error

  def DecodeSerializationTag(self) -> Tuple[int, definitions.SerializationTag]:
    """Decodes a SerializationTag.

    Returns:
      a tuple of the offset and the serialization tag.

    Raises:
      ParserError if an invalid terminal value was encountered.
    """
    offset, terminal_byte = self.decoder.DecodeUint8()
    try:
      return offset, definitions.SerializationTag(terminal_byte)
    except ValueError as error:
      raise errors.ParserError(
          f'Invalid terminal {terminal_byte} at offset {offset}') from error

  def DecodeArray(self) -> JSArray:
    """Decodes an Array value.

    Returns:
      the JavaScript array.

    Raises:
      ParserError if an invalid Terminator tag was found.
    """
    _, length = self.decoder.DecodeUint32()
    array = JSArray()
    for _ in range(length):
      _, _ = self.decoder.DecodeUint32()
      _, value = self.DecodeValue()
      array.append(value)

    offset, terminator_tag = self.decoder.DecodeUint32()
    if terminator_tag != definitions.TerminatorTag:
      raise errors.ParserError(f'Terminator tag not found at offset {offset}.')

    offset, tag = self.decoder.DecodeUint32()
    if tag == definitions.NonIndexPropertiesTag:
      while tag != definitions.TerminatorTag:
        name = self.DecodeStringData()
        _, value = self.DecodeValue()
        _, tag = self.decoder.DecodeUint32()
        array.properties[name] = value
    elif tag != definitions.TerminatorTag:
      raise errors.ParserError(f'Terminator tag not found at offset {offset}.')
    return array

  def DecodeObject(self) -> Dict[str, Any]:
    """Decodes an Object value."""
    tag = self.PeekTag()
    js_object = {}
    while tag != definitions.TerminatorTag:
      name = self.DecodeStringData()
      _, value = self.DecodeValue()
      js_object[name] = value
      tag = self.PeekTag()
    _ = self.decoder.DecodeUint32()
    self.object_pool.append(js_object)
    return js_object

  def DecodeStringData(self) -> str:
    """Decodes a StringData value.

    Returns:
      A JavaScript array.

    Raises:
      ParserError if an:
          * unexpected TerminatorTag is found
          * unexpected constant pool size value is found
          * disallowed string length is found.
          * unable to to decode a buffer as utf-16-le.
    """
    peeked_tag = self.PeekTag()
    if peeked_tag == definitions.TerminatorTag:
      raise errors.ParserError('Unexpected TerminatorTag found')

    if peeked_tag == definitions.StringPoolTag:
      _ = self.decoder.DecodeUint32()
      if len(self.constant_pool) < 0xff:
        _, cp_index = self.decoder.DecodeUint8()
      elif len(self.constant_pool) < 0xffff:
        _, cp_index = self.decoder.DecodeUint16()
      elif len(self.constant_pool) < 0xffffffff:
        _, cp_index = self.decoder.DecodeUint32()
      else:
        raise errors.ParserError('Unexpected constant pool size value.')
      return self.constant_pool[cp_index]

    _, length_with_8bit_flag = self.decoder.DecodeUint32()
    if length_with_8bit_flag == definitions.TerminatorTag:
      raise errors.ParserError('Disallowed string length found.')

    length = length_with_8bit_flag & 0x7FFFFFFF
    is_8bit = length_with_8bit_flag & definitions.StringDataIs8BitFlag

    if is_8bit:
      _, characters = self.decoder.ReadBytes(length)
      value = characters.decode('latin-1')
    else:
      _, characters = self.decoder.ReadBytes(2*length)
      try:
        value = characters.decode('utf-16-le')
      except UnicodeDecodeError:
        raise errors.ParserError(
            f'Unable to decode {len(characters)} characters as utf-16-le')
    self.constant_pool.append(value)
    return value

  def DecodeDate(self) -> datetime:
    """Decodes a Date value."""
    _, timestamp = self.decoder.DecodeDouble()
    value = datetime.utcfromtimestamp(timestamp/1000)
    return value

  def DecodeFileData(self) -> FileData:
    """Decodes a FileData value."""
    path = self.DecodeStringData()
    url = self.DecodeStringData()
    file_type = self.DecodeStringData()
    name = self.DecodeStringData()
    _, last_modified = self.decoder.DecodeDouble()

    return FileData(
        path=path,
        url=url,
        type=file_type,
        name=name,
        last_modified=last_modified)

  def DecodeFileList(self) -> FileList:
    """Decodes a FileList value."""
    _, length = self.decoder.DecodeUint32()
    file_list = []
    for _ in range(length):
      file_list.append(self.DecodeFileData())
    return FileList(files=file_list)

  def DecodeImageData(self) -> Dict[str, Any]:
    """Decodes an ImageData value."""
    _, width = self.decoder.DecodeUint32()
    _, height = self.decoder.DecodeUint32()
    _, length = self.decoder.DecodeUint32()
    data = self.decoder.ReadBytes(length)

    if self.version > 7:
      _, color_space = self.decoder.DecodeUint8()
    else:
      color_space = None

    # TODO: make this a dataclass?
    return {
      'width': width,
      'height': height,
      'length': length,
      'data': data,
      'color_space': color_space
    }

  def DecodeBlob(self) -> Dict[str, Any]:
    """Decodes a Blob value."""
    url = self.DecodeStringData()
    blob_type = self.DecodeStringData()
    size = self.decoder.DecodeUint64()
    if self.version >= 11:
      _, memory_cost = self.decoder.DecodeUint64()
    else:
      memory_cost = None

    # TODO: make this a dataclass?
    return {
      'url': url,
      'blob_type': blob_type,
      'size': size,
      'memory_cost': memory_cost
    }

  def DecodeRegExp(self) -> RegExp:
    """Decodes a RegExp value."""
    pattern = self.DecodeStringData()
    flags = self.DecodeStringData()
    return RegExp(pattern=pattern, flags=flags)

  def DecodeMapData(self) -> dict:
    """Decodes a Map value."""
    tag = self.PeekSerializationTag()
    js_map = {}   # TODO: make this into a JSMap (like JSArray/JSSet)

    while tag != definitions.SerializationTag.NON_MAP_PROPERTIES:
      _, key = self.DecodeValue()
      _, value = self.DecodeValue()
      js_map[key] = value
      tag = self.PeekSerializationTag()

    # consume the NonMapPropertiesTag
    _, tag = self.DecodeSerializationTag()

    pool_tag = self.PeekTag()
    while pool_tag != definitions.TerminatorTag:
      name = self.DecodeStringData()
      value = self.DecodeValue()
      js_map[name] = value
      pool_tag = self.PeekTag()

    _, tag = self.decoder.DecodeUint32()

    return js_map

  def DecodeSetData(self) -> JSSet:
    """Decodes a SetData value."""
    tag = self.PeekSerializationTag()
    js_set = JSSet()

    while tag != definitions.SerializationTag.NON_SET_PROPERTIES:
      _, key = self.DecodeValue()
      js_set.add(key)
      tag = self.PeekSerializationTag()

    # consume the NonSetPropertiesTag
    _, tag = self.DecodeSerializationTag()

    pool_tag = self.PeekTag()
    while pool_tag != definitions.TerminatorTag:
      name = self.DecodeStringData()
      value = self.DecodeValue()
      js_set.properties[name] = value
      pool_tag = self.decoder.PeekBytes(4)

    # consume the TerminatorTag
    _, tag = self.decoder.DecodeUint32()
    return js_set

  def DecodeCryptoKey(self) -> bytes:
    """Decodes a CryptoKey value."""
    _, wrapped_key_length = self.decoder.DecodeUint32()
    _, wrapped_key = self.decoder.ReadBytes(wrapped_key_length)
    key = plistlib.loads(wrapped_key)  # TODO: unwrap the wrapped key.
    return key

  def DecodeBigIntData(self) -> int:
    """Decodes a BigIntData value."""
    _, sign = self.decoder.DecodeUint8()
    _, number_of_elements = self.decoder.DecodeUint32()
    contents = []
    for _ in range(number_of_elements):
      _, element = self.decoder.ReadBytes(8)
      contents.extend(element)
    value = int.from_bytes(contents, byteorder='little', signed=bool(sign))
    return value

  def DecodeArrayBuffer(self) -> bytes:
    """Decodes an ArrayBuffer value."""
    _, byte_length = self.decoder.DecodeUint64()
    _, buffer = self.decoder.ReadBytes(byte_length)
    self.object_pool.append(buffer)
    return buffer

  def DecodeResizableArrayBuffer(self) -> ResizableArrayBuffer:
    """Decodes an ArrayBuffer value."""
    _, byte_length = self.decoder.DecodeUint64()
    _, max_length = self.decoder.DecodeUint64()
    _, buffer = self.decoder.ReadBytes(byte_length)
    self.object_pool.append(buffer)
    return ResizableArrayBuffer(buffer=buffer, max_length=max_length)

  def DecodeArrayBufferTransfer(self) -> int:
    """Decodes an ArrayBufferTransfer value."""
    _, value = self.decoder.DecodeUint32()
    return value

  def DecodeSharedArrayBuffer(self) -> int:
    """Decodes an SharedArrayBuffer value."""
    _, value = self.decoder.DecodeUint32()
    return value

  def DecodeObjectReference(self) -> Any:
    """Decodes an ObjectReference value."""
    _, object_ref = self.decoder.DecodeUint8()
    return self.object_pool[object_ref - 1]

  def DecodeArrayBufferView(self) -> ArrayBufferView:
    """Decodes an ArrayBufferView value.

    Returns:
      an ArrayBufferView.

    Raises:
      ParserError if an unexpected serialization tag is found.
    """
    _, array_buffer_view_subtag = self.decoder.DecodeUint8()
    array_buffer_view_subtag = definitions.ArrayBufferViewSubtag(
        array_buffer_view_subtag)
    _, byte_offset = self.decoder.DecodeUint64()
    _, byte_length = self.decoder.DecodeUint64()
    _, next_serialization_tag = self.DecodeSerializationTag()

    if next_serialization_tag == definitions.SerializationTag.ARRAY_BUFFER:
      value = self.DecodeArrayBuffer()
    elif (next_serialization_tag ==
          definitions.SerializationTag.OBJECT_REFERENCE):
      value = self.DecodeObjectReference()
    else:
      raise errors.ParserError(
          f'Unexpected serialization tag {next_serialization_tag}.')
    return ArrayBufferView(
        array_buffer_view_subtag=array_buffer_view_subtag,
        buffer=value,
        offset=byte_offset,
        length=byte_length)

  def DecodeSerializedValue(self) -> Any:
    """Decodes a serialized value.

    Returns:
      the serialized value.

    Raises:
      ParserError when CurrentVersion is not found.
    """
    _, current_version = self.decoder.DecodeUint32()
    if current_version != definitions.CurrentVersion:
      raise errors.ParserError(
          f'{current_version} is not the expected CurrentVersion')
    _, value = self.DecodeValue()
    return value

  def DecodeValue(self) -> Tuple[int, Any]:
    """Decodes a value.

    Returns:
      the offset and parsed value.

    Raises:
      ParserError when an unhandled SerializationTag is found.
    """
    offset, tag = self.DecodeSerializationTag()
    if tag == definitions.SerializationTag.ARRAY:
      value = self.DecodeArray()
    elif tag == definitions.SerializationTag.OBJECT:
      value = self.DecodeObject()
    elif tag == definitions.SerializationTag.UNDEFINED:
      value = Undefined()
    elif tag == definitions.SerializationTag.NULL:
      value = Null()
    elif tag == definitions.SerializationTag.INT:
      _, value = self.decoder.DecodeInt32()
    elif tag == definitions.SerializationTag.ZERO:
      value = 0
    elif tag == definitions.SerializationTag.ONE:
      value = 1
    elif tag == definitions.SerializationTag.FALSE:
      value = False
    elif tag == definitions.SerializationTag.TRUE:
      value = True
    elif tag == definitions.SerializationTag.DOUBLE:
      _, value = self.decoder.DecodeDouble()
    elif tag == definitions.SerializationTag.DATE:
      value = self.DecodeDate()
    elif tag == definitions.SerializationTag.FILE:
      value = self.DecodeFileData()
    elif tag == definitions.SerializationTag.FILE_LIST:
      value = self.DecodeFileList()
    elif tag == definitions.SerializationTag.IMAGE_DATA:
      value = self.DecodeImageData()
    elif tag == definitions.SerializationTag.BLOB:
      value = self.DecodeBlob()
    elif tag == definitions.SerializationTag.STRING:
      value = self.DecodeStringData()
    elif tag == definitions.SerializationTag.EMPTY_STRING:
      value = ''
    elif tag == definitions.SerializationTag.REG_EXP:
      value = self.DecodeRegExp()
    elif tag == definitions.SerializationTag.OBJECT_REFERENCE:
      value = self.DecodeObjectReference()
    elif tag == definitions.SerializationTag.ARRAY_BUFFER:
      value = self.DecodeArrayBuffer()
    elif tag == definitions.SerializationTag.ARRAY_BUFFER_VIEW:
      value = self.DecodeArrayBufferView()
    elif tag == definitions.SerializationTag.ARRAY_BUFFER_TRANSFER:
      value = self.DecodeArrayBufferTransfer()
    elif tag == definitions.SerializationTag.TRUE_OBJECT:
      self.object_pool.append(True)
      value = True
    elif tag == definitions.SerializationTag.FALSE_OBJECT:
      self.object_pool.append(False)
      value = False
    elif tag == definitions.SerializationTag.STRING_OBJECT:
      value = self.DecodeStringData()
      self.object_pool.append(value)
    elif tag == definitions.SerializationTag.EMPTY_STRING_OBJECT:
      value = ''
      self.object_pool.append(value)
    elif tag == definitions.SerializationTag.NUMBER_OBJECT:
      _, value = self.decoder.DecodeDouble()
      self.object_pool.append(value)
    elif tag == definitions.SerializationTag.SET_OBJECT:
      value = self.DecodeSetData()
    elif tag == definitions.SerializationTag.MAP_OBJECT:
      value = self.DecodeMapData()
    elif tag == definitions.SerializationTag.CRYPTO_KEY:
      value = self.DecodeCryptoKey()
    elif tag == definitions.SerializationTag.SHARED_ARRAY_BUFFER:
      value = self.DecodeSharedArrayBuffer()
    elif tag == definitions.SerializationTag.BIGINT:
      value = self.DecodeBigIntData()
    elif tag == definitions.SerializationTag.BIGINT_OBJECT:
      value = self.DecodeBigIntData()
      self.object_pool.append(value)
    else:
      raise errors.ParserError(f'Unhandled Serialization Tag {tag.name} found.')
    return offset, value

  @classmethod
  def FromBytes(cls, data: bytes) -> Any:
    """Returns a deserialized JavaScript object from the data.

    Args:
      data: the data to deserialize/parse.

    Returns:
      A python representation of the parsed JavaScript object.

    Raises:
      errors.ParserError: if there is an invalid V8 JavaScript header.
    """
    stream = io.BytesIO(data)
    deserializer = cls(stream)
    return deserializer.DecodeSerializedValue()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Safari IndexedDB records."""
from dataclasses import dataclass
import plistlib
import sqlite3
import sys
import traceback
from typing import Any, Generator, Optional

from dfindexeddb import errors
from dfindexeddb.indexeddb.safari import webkit


@dataclass
class ObjectStoreInfo:
  """An ObjectStoreInfo.

  Attributes:
    id: the object store ID.
    name: the object store name.
    key_path: the object store key path.
    auto_inc: True if the object store uses auto incrementing IDs.
    database_name: the database name from the IDBDatabaseInfo table.
  """
  id: int
  name: str
  key_path: str
  auto_inc: bool
  database_name: str


@dataclass
class IndexedDBRecord:
  """A Safari IndexedDBRecord.

  Attributes:
    key: the parsed key.
    value: the parsed value.
    object_store_id: the object store id.
    object_store_name: the object store name from the ObjectStoreInfo table.
    database_name: the IndexedDB database name from the IDBDatabaseInfo table.
    record_id: the record ID from the Record table.
  """
  key: Any
  value: Any
  object_store_id: int
  object_store_name: str
  database_name: str
  record_id: int


class FileReader:
  """A reader for Safari IndexedDB sqlite3 files.

  Attributes:
    database_name: the database name.
    database_version: the database version.
    metadata_version: the metadata version.
    max_object_store_id: the maximum object store ID.
  """

  def __init__(self, filename: str):
    """Initializes the FileReader.

    Args:
      filename: the IndexedDB filename.
    """
    self.filename = filename

    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "DatabaseVersion"')
      result = cursor.fetchone()
      self.database_version = result[0]

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "MetadataVersion"')
      result = cursor.fetchone()
      self.metadata_version = result[0]

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "DatabaseName"')
      result = cursor.fetchone()
      self.database_name = result[0]

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "MaxObjectStoreID"')
      result = cursor.fetchone()
      self.max_object_store_id = result[0]

  def ObjectStores(self) -> Generator[ObjectStoreInfo, None, None]:
    """Returns the Object Store information from the IndexedDB database.

    Yields:
      ObjectStoreInfo instances.
    """
    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      cursor = conn.execute(
          'SELECT id, name, keypath, autoinc FROM ObjectStoreInfo')
      results = cursor.fetchall()
      for result in results:
        key_path = plistlib.loads(result[2])
        yield ObjectStoreInfo(
            id=result[0],
            name=result[1],
            key_path=key_path,
            auto_inc=result[3],
            database_name=self.database_name)

  def RecordById(self, record_id: int) -> Optional[IndexedDBRecord]:
    """Returns an IndexedDBRecord for the given record_id.

    Returns:
      the IndexedDBRecord or None if the record_id does not exist in the 
          database.
    """
    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          'SELECT r.key, r.value, r.objectStoreID, o.name, r.recordID FROM '
          'Records r '
          'JOIN ObjectStoreInfo o ON r.objectStoreID == o.id '
          'WHERE r.recordID = ?', (record_id, ))
      row = cursor.fetchone()
      if not row:
        return None
      key = webkit.IDBKeyData.FromBytes(row[0]).data
      value = webkit.SerializedScriptValueDecoder.FromBytes(row[1])
      return IndexedDBRecord(
          key=key,
          value=value,
          object_store_id=row[2],
          object_store_name=row[3].decode('utf-8'),
          database_name=self.database_name,
          record_id=row[4])

  def RecordsByObjectStoreName(
      self,
      name: str
  ) -> Generator[IndexedDBRecord, None, None]:
    """Returns IndexedDBRecords for the given ObjectStore name.

    Yields:
      IndexedDBRecord instances.
    """
    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      conn.text_factory = bytes
      for row in conn.execute(
          'SELECT r.key, r.value, r.objectStoreID, o.name, r.recordID FROM '
          'Records r '
          'JOIN ObjectStoreInfo o ON r.objectStoreID == o.id '
          'WHERE o.name = ?', (name, )):
        key = webkit.IDBKeyData.FromBytes(row[0]).data
        value = webkit.SerializedScriptValueDecoder.FromBytes(row[1])
        yield IndexedDBRecord(
            key=key,
            value=value,
            object_store_id=row[2],
            object_store_name=row[3].decode('utf-8'),
            database_name=self.database_name,
            record_id=row[4])

  def RecordsByObjectStoreId(
      self,
      object_store_id: int
  ) -> Generator[IndexedDBRecord, None, None]:
    """Returns IndexedDBRecords for the given ObjectStore id.

    Yields:
      IndexedDBRecord instances.
    """
    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          'SELECT r.key, r.value, r.objectStoreID, o.name, r.recordID '
          'FROM Records r '
          'JOIN ObjectStoreInfo o ON r.objectStoreID == o.id '
          'WHERE o.id = ?', (object_store_id, ))
      for row in cursor:
        key = webkit.IDBKeyData.FromBytes(row[0]).data
        value = webkit.SerializedScriptValueDecoder.FromBytes(row[1])
        yield IndexedDBRecord(
            key=key,
            value=value,
            object_store_id=row[2],
            object_store_name=row[3].decode('utf-8'),
            database_name=self.database_name,
            record_id=row[4])

  def Records(self) -> Generator[IndexedDBRecord, None, None]:
    """Returns all the IndexedDBRecords."""
    with sqlite3.connect(f'file:{self.filename}?mode=ro', uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          'SELECT r.key, r.value, r.objectStoreID, o.name, r.recordID '
          'FROM Records r '
          'JOIN ObjectStoreInfo o ON r.objectStoreID == o.id')
      for row in cursor:
        try:
          key = webkit.IDBKeyData.FromBytes(row[0]).data
        except(
            errors.ParserError,
            errors.DecoderError,
            NotImplementedError) as err:
          print(
              f'Error parsing IndexedDB key: {err}', file=sys.stderr)
          print(f'Traceback: {traceback.format_exc()}', file=sys.stderr)
          continue
        try:
          value = webkit.SerializedScriptValueDecoder.FromBytes(row[1])
        except(
            errors.ParserError,
            errors.DecoderError,
            NotImplementedError) as err:
          print(
              f'Error parsing IndexedDB value: {err}', file=sys.stderr)
          print(f'Traceback: {traceback.format_exc()}', file=sys.stderr)
          continue
        yield IndexedDBRecord(
            key=key,
            value=value,
            object_store_id=row[2],
            object_store_name=row[3].decode('utf-8'),
            database_name=self.database_name,
            record_id=row[4])



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser for LevelDB Descriptor (MANIFEST) files."""
from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Generator, Optional

from dfindexeddb import errors
from dfindexeddb.leveldb import definitions
from dfindexeddb.leveldb import log
from dfindexeddb.leveldb import utils


@dataclass
class InternalKey:
  """An InternalKey.

  Attributes:
    offset: the offset.
    user_key: the user key.
    sequence_number: the sequence number.
    key_type: the key type.
  """
  offset: int
  user_key: bytes
  sequence_number: int
  key_type: int

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> InternalKey:
    """Decodes an InternalKey from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The InternalKey instance.
    """
    offset, slice_bytes = decoder.DecodeLengthPrefixedSlice()

    if len(slice_bytes) < definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH:
      raise errors.ParserError('Insufficient bytes to parse InternalKey')

    user_key = slice_bytes[:-definitions.SEQUENCE_LENGTH]
    sequence_number = int.from_bytes(
        slice_bytes[-definitions.SEQUENCE_LENGTH:],
        byteorder='little',
        signed=False)
    key_type = slice_bytes[-definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH]

    return cls(
        offset=base_offset + offset,
        user_key=user_key,
        sequence_number=sequence_number,
        key_type=key_type)


@dataclass
class NewFile(utils.FromDecoderMixin):
  """A NewFile.

  Attributes:
    offset: the offset.
    level: the level.
    number: the file number.
    file_size: the file size.
    smallest: the smallest internal key.
    largest: the largest internal key.
  """
  offset: int
  level: int
  number: int
  file_size: int
  smallest: InternalKey
  largest: InternalKey

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> NewFile:
    """Decodes a NewFile from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The NewFile instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, number = decoder.DecodeUint64Varint()
    _, file_size = decoder.DecodeUint64Varint()
    smallest = InternalKey.FromDecoder(decoder, base_offset=base_offset)
    largest = InternalKey.FromDecoder(decoder, base_offset=base_offset)

    return cls(
        offset=offset + base_offset,
        level=level,
        number=number,
        file_size=file_size,
        smallest=smallest,
        largest=largest)


@dataclass
class CompactPointer(utils.FromDecoderMixin):
  """A CompactPointer.

  Attributes:
    offset: the offset.
    level: the level.
    key: the key bytes.
  """
  offset: int
  level: int
  key: bytes

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0
  ) -> CompactPointer:
    """Decodes a CompactPointer from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The CompactPointer instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, key = decoder.DecodeLengthPrefixedSlice()
    return cls(offset=base_offset + offset, level=level, key=key)


@dataclass
class DeletedFile(utils.FromDecoderMixin):
  """A DeletedFile.

  Attributes:
    offset: the offset.
    level: the level.
    number: the file number.
  """
  offset: int
  level: int
  number: int

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> DeletedFile:
    """Decodes a DeletedFile from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The DeletedFile instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, number = decoder.DecodeUint64Varint()
    return cls(offset=base_offset + offset, level=level, number=number)


@dataclass
class VersionEdit(utils.FromDecoderMixin):
  """A VersionEdit is recorded in a LevelDB descriptor/manifest file.

  Attributes:
    offset: the offset where the VersionEdit was parsed.
    comparator: the comparator.
    log_number: the log number.
    prev_log_number: the previous log number.
    next_file_number: the next file number.
    last_sequence: the last sequence.
    compact_pointers: the list of CompactPointers.
    deleted_files: the list of DeletedFiles.
    new_files: the list of NewFiles.
  """
  offset: int
  comparator: Optional[bytes] = None
  log_number: Optional[int] = None
  prev_log_number: Optional[int] = None
  next_file_number: Optional[int] = None
  last_sequence: Optional[int] = None
  compact_pointers: list[CompactPointer] = field(default_factory=list)
  deleted_files: list[DeletedFile] = field(default_factory=list)
  new_files: list[NewFile] = field(default_factory=list)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> VersionEdit:
    """Decodes a VersionEdit from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The VersionEdit instance.

    Raises:
      ParserError if an invalid VersionEditTag is parsed.
    """
    offset, tag_byte = decoder.DecodeUint32Varint()
    version_edit = cls(offset=base_offset + offset)

    while tag_byte:
      try:
        tag = definitions.VersionEditTags(tag_byte)
      except TypeError as error:
        raise errors.ParserError(
            f'Invalid VersionEditTag at offset {offset}') from error

      if tag == definitions.VersionEditTags.COMPARATOR:
        _, version_edit.comparator = decoder.DecodeLengthPrefixedSlice()
      elif tag == definitions.VersionEditTags.LOG_NUMBER:
        _, version_edit.log_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.PREV_LOG_NUMBER:
        _, version_edit.prev_log_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.NEXT_FILE_NUMBER:
        _, version_edit.next_file_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.LAST_SEQUENCE:
        _, version_edit.last_sequence = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.COMPACT_POINTER:
        compact_pointer = CompactPointer.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.compact_pointers.append(compact_pointer)
      elif tag == definitions.VersionEditTags.DELETED_FILE:
        deleted_file = DeletedFile.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.deleted_files.append(deleted_file)
      elif tag == definitions.VersionEditTags.NEW_FILE:
        file_metadata = NewFile.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.new_files.append(file_metadata)

      if decoder.NumRemainingBytes() == 0:
        break

      offset, tag_byte = decoder.DecodeUint32Varint()

    return version_edit


@dataclass
class LevelDBVersion:
  """A LevelDBVersion.

  A LevelDBVersion represents the current state of the table files and log file
  that are currently "active".  The current state is determined by the sequence
  of VersionEdits that are parsed from a descriptor file.

  Active files can contain overlapping keys in the current log file and the
  "young" or 0-level.

  "Deleted files" will typically no longer exist but may be forensically
  recoverable.
  """
  current_log: str
  version_edit_offset: int
  last_sequence: int
  active_files: dict[int, dict[int, NewFile]]
  deleted_files: dict[int, dict[int, DeletedFile]]


class FileReader:
  """A reader for Descriptor files.

  A DescriptorFileReader provides read-only sequential iteration of serialized
  structures in a leveldb Descriptor file.  These structures include:
  * blocks (Block)
  * records (PhysicalRecord)
  * version edits (VersionEdit)
  """
  def __init__(self, filename: str):
    """Initializes the Descriptor a.k.a. MANIFEST file.

    Args:
      filename: the Descriptor filename (e.g. MANIFEST-000001)
    """
    self.filename = filename


  def GetBlocks(self) -> Generator[log.Block, None, None]:
    """Returns an iterator of Block instances.

    A Descriptor file is composed of one or more blocks.

    Yields:
      Block
    """
    with open(self.filename, 'rb') as fh:
      block = log.Block.FromStream(fh)
      while block:
        yield block
        block = log.Block.FromStream(fh)

  def GetPhysicalRecords(self) -> Generator[log.PhysicalRecord, None, None]:
    """Returns an iterator of PhysicalRecord instances.

    A block is composed of one or more physical records.

    Yields:
      PhysicalRecord
    """
    for block in self.GetBlocks():
      yield from block.GetPhysicalRecords()

  def GetVersionEdits(self) -> Generator[VersionEdit, None, None]:
    """Returns an iterator of VersionEdit instances.

    Depending on the VersionEdit size, it can be spread across one or
    more physical records.

    Yields:
      VersionEdit
    """
    buffer = bytearray()
    for physical_record in self.GetPhysicalRecords():
      if (physical_record.record_type ==
          definitions.LogFilePhysicalRecordType.FULL):
        buffer = physical_record.contents
        offset = physical_record.contents_offset + physical_record.base_offset
        version_edit = VersionEdit.FromBytes(buffer, base_offset=offset)
        yield version_edit
        buffer = bytearray()
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.FIRST):
        offset = physical_record.contents_offset + physical_record.base_offset
        buffer = bytearray(physical_record.contents)
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.MIDDLE):
        buffer.extend(bytearray(physical_record.contents))
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.LAST):
        buffer.extend(bytearray(physical_record.contents))
        version_edit = VersionEdit.FromBytes(buffer, base_offset=offset)
        yield version_edit
        buffer = bytearray()

  def GetVersions(self) -> Generator[LevelDBVersion, None, None]:
    """Returns an iterator of LevelDBVersion instances.

    Yields:
      LevelDBVersion
    """
    active_files = defaultdict(dict)
    deleted_files = defaultdict(set)
    current_log = None

    for version_edit in self.GetVersionEdits():
      if version_edit.log_number:
        current_log = f'{version_edit.log_number:06d}.log'

      for new_file in version_edit.new_files:
        active_files[new_file.level][f'{new_file.number:06d}.ldb'] = new_file

      for deleted_file in version_edit.deleted_files:
        active_files[deleted_file.level].pop(f'{deleted_file.number:06d}.ldb')
        deleted_files[deleted_file.level].add(f'{deleted_file.number:06d}.ldb')

      yield LevelDBVersion(
          current_log=current_log,
          active_files=dict(active_files),
          deleted_files=dict(deleted_files),
          version_edit_offset=version_edit.offset,
          last_sequence=version_edit.last_sequence)

  def GetLatestVersion(self) -> LevelDBVersion:
    """Returns the latest LevelDBVersion instance."""
    *_, latest = self.GetVersions()
    return latest



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for blink javascript serialized objects."""
from __future__ import annotations
from dataclasses import dataclass
import io
from typing import Any, Optional, Union

from dfindexeddb import errors
from dfindexeddb import utils
from dfindexeddb.indexeddb.chromium import definitions
from dfindexeddb.indexeddb.chromium import v8


_MS_PER_SECOND = 1000


@dataclass
class AudioData:
  """A Javascript AudioData class.

  Attributes:
    audio_frame_index: the Audio Frame index.
  """
  audio_frame_index: int


@dataclass
class BaseIndex:
  """A base index class."""
  index: int


@dataclass
class BitMap:
  """A Javascript BitMap."""


@dataclass
class Blob:
  """A Javascript Blob class type.

  Attributes:
    uuid: the UUID of the Blob.
    type: the type of the Blob.
    size: the size of the Blob.
  """
  uuid: str
  type: str
  size: int


@dataclass
class BlobIndex(BaseIndex):
  """A Javascript BlobIndex class type."""


@dataclass
class CryptoKey:
  """A Javascript CryptoKey class type.

  Attributes:
    algorithm: the CryptoKey algorithm.
    key_type: the CryptoKey type.
    extractable: True if the CryptoKey is extractable.
    usages: the CryptoKey usage.
    key_data: the raw key data.
  """
  algorithm_parameters: dict[str, Any]
  key_type: Union[
      definitions.WebCryptoKeyType, definitions.AsymmetricCryptoKeyType]
  extractable: bool
  usages: definitions.CryptoKeyUsage
  key_data: bytes


@dataclass
class DOMException:
  """A Javascript DOMException.

  Attributes:
    name: the name.
    message: the message.
    stack_unused: the stack unused.
  """
  name: str
  message: str
  stack_unused: str


@dataclass
class DOMFileSystem:
  """A Javascript DOMFileSystem.

  Attributes:
    raw_type: the raw type.
    name: the name.
    root_url: the root URL.
  """
  raw_type: int
  name: str
  root_url: str


@dataclass
class DOMMatrix2D:
  """A Javascript DOMMatrix2D.

  Attributes:
    values: the values.
  """
  values: list[float]


@dataclass
class DOMMatrix2DReadOnly(DOMMatrix2D):
  """A Javascript Read-Only DOMMatrix2D."""


@dataclass
class DOMMatrix:
  """A Javascript DOMMatrix.

  Attributes:
    values: the values.
  """
  values: list[float]


@dataclass
class DOMMatrixReadOnly(DOMMatrix):
  """A Javascript Read-Only DOMMatrix."""


@dataclass
class DOMPoint:
  """A Javascript DOMPoint.

  Attributes:
    x: the X coordinate.
    y: the Y coordinate.
    z: the Z coordinate.
    w: the W coordinate.
  """
  x: float
  y: float
  z: float
  w: float


@dataclass
class DOMPointReadOnly(DOMPoint):
  """A Javascript Read-Only DOMPoint."""


@dataclass
class DOMQuad:
  """A Javascript DOMQuad.

  Attributes:
    p1: the first point.
    p2: the second point.
    p3: the third point.
    p4: the fourth point.
  """
  p1: DOMPoint
  p2: DOMPoint
  p3: DOMPoint
  p4: DOMPoint


@dataclass
class DOMRect:
  """A Javascript DOMRect.

  Attributes:
    x: the X coordinate.
    y: the Y coordinate.
    width: the width.
    height: the height.
  """
  x: float
  y: float
  width: float
  height: float


@dataclass
class DOMRectReadOnly(DOMRect):
  """A Javascript Read-only DOMRect."""


@dataclass
class EncodedAudioChunk(BaseIndex):
  """A Javascript EncodedAudioChunk."""


@dataclass
class EncodedVideoChunk(BaseIndex):
  """A Javascript EncodedVideoChunk."""


@dataclass
class File:
  """A Javascript File.

  Attributes:
    path: the file path.
    name: the file name.
    relative_path: the file relative path.
    uuid: the file UUID.
    type: the file type.
    has_snapshot: True if the file has a snapshot.
    size: the file size.
    last_modified_ms: the file last modified in milliseconds.
    is_user_visible: True if the file is user visible.
  """
  path: str
  name: Optional[str]
  relative_path: Optional[str]
  uuid: str
  type: str
  has_snapshot: int
  size: Optional[int]
  last_modified_ms: Optional[float]
  is_user_visible: int


@dataclass
class FileIndex(BaseIndex):
  """A Javascript FileIndex."""


@dataclass
class FileList:
  """A Javascript FileList.

  Attributes:
    files: the list of files.
  """
  files: list[File]


@dataclass
class FileListIndex:
  """A Javascript FileListIndex.

  Attributes:
    file_indexes: the list of file indexes.
  """
  file_indexes: list[FileIndex]


@dataclass
class FileSystemHandle:
  """A Javascript FileSystemHandle.

  Attributes:
    name: the file system handle name.
    token_index: the file system token index.
  """
  name: str
  token_index: int


@dataclass
class ImageBitmapTransfer(BaseIndex):
  """A Javascript ImageBitmapTransfer."""


@dataclass
class MediaSourceHandle(BaseIndex):
  """A Javascript MediaSourceHandle."""


@dataclass
class MessagePort(BaseIndex):
  """A Javascript MessagePort."""


@dataclass
class MojoHandle(BaseIndex):
  """A Javascript MojoHandle."""


@dataclass
class ReadableStreamTransfer(BaseIndex):
  """A Javascript ReadableStreamTransfer."""


@dataclass
class RTCEncodedAudioFrame(BaseIndex):
  """A Javascript RTCEncodedAudioFrame."""


@dataclass
class RTCEncodedVideoFrame(BaseIndex):
  """A Javascript RTCEncodedVideoFrame."""


@dataclass
class WriteableStreamTransfer(BaseIndex):
  """A Javascript WriteableStreamTransfer."""


@dataclass
class TransformStreamTransfer(BaseIndex):
  """A Javascript TransformStreamTransfer."""


@dataclass
class OffscreenCanvasTransfer:
  """A Javascript Offscreen Canvas Transfer."""
  width: int
  height: int
  canvas_id: int
  client_id: int
  sink_id: int
  filter_quality: int


@dataclass
class UnguessableToken:
  """A Javascript Unguessable Token.

  Attributes:
    high: the high part.
    low: the low part.
  """
  high: int
  low: int


@dataclass
class VideoFrame(BaseIndex):
  """A Javascript VideoFrame."""


class V8ScriptValueDecoder:
  """A Blink V8 Serialized Script Value (SSV) decoder.

  Attributes:
    deserializer (v8.ValueDeserializer): the V8 value deserializer.
    raw_data (bytes): the raw bytes containing the serialized javascript
        value.
    version (int): the blink version.
  """
  _MIN_VERSION_FOR_SEPARATE_ENVELOPE = 16

  # As defined in trailer_reader.h
  _MIN_WIRE_FORMAT_VERSION = 21

  def __init__(self, raw_data: bytes):
    self.deserializer = None
    self.raw_data = raw_data
    self.version = 0
    self.trailer_offset = None
    self.trailer_size = None

  def _ReadVersionEnvelope(self) -> int:
    """Reads the Blink version envelope.

    Returns:
      The number of bytes consumed reading the Blink version envelope or zero
          if no blink envelope is detected.
    """
    if not self.raw_data:
      return 0

    decoder = utils.StreamDecoder(io.BytesIO(self.raw_data))
    _, tag = decoder.DecodeUint8()
    if tag != definitions.BlinkSerializationTag.VERSION:
      return 0

    _, self.version = decoder.DecodeUint32Varint()
    if self.version < self._MIN_VERSION_FOR_SEPARATE_ENVELOPE:
      return 0

    consumed_bytes = decoder.stream.tell()

    if self.version >= self._MIN_WIRE_FORMAT_VERSION:
      _, trailer_offset_tag = decoder.DecodeUint8()
      if (trailer_offset_tag !=
          definitions.BlinkSerializationTag.TRAILER_OFFSET):
        raise errors.ParserError('Trailer offset not found')
      _, self.trailer_offset = decoder.DecodeInt(
          byte_count=8, byte_order='big', signed=False)
      _, self.trailer_size = decoder.DecodeInt(
          byte_count=4, byte_order='big', signed=False)
      trailer_offset_data_size = 1 + 8 + 4 # 1 + sizeof(uint64) + sizeof(uint32)
      consumed_bytes += trailer_offset_data_size
      if consumed_bytes >= len(self.raw_data):
        return 0
    return consumed_bytes

  def _ReadAESKey(self) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads an AES CryptoKey.

    Returns:
      The AES key algorithm parameters
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm_id = definitions.CryptoKeyAlgorithm(raw_id)

    _, length_bytes = self.deserializer.decoder.DecodeUint32Varint()

    algorithm_parameters = {
        'id': crypto_key_algorithm_id,
        'length_bits': length_bytes*8
    }

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadHMACKey(self) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads a HMAC CryptoKey.

    Returns:
      The HMAC key algorithm parameters
    """
    _, length_bytes = self.deserializer.decoder.DecodeUint32Varint()
    _, raw_hash = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_hash)

    algorithm_parameters = {
        'id': crypto_key_algorithm,
        'length_bits': length_bytes*8
    }

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadRSAHashedKey(
      self
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads an RSA Hashed CryptoKey.

    Returns:
      The RSA Hashed key algorithm parameters
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(raw_key_type)

    _, modulus_length_bits = self.deserializer.decoder.DecodeUint32Varint()
    _, public_exponent_size = self.deserializer.decoder.DecodeUint32Varint()
    _, public_exponent_bytes = self.deserializer.decoder.ReadBytes(
        count=public_exponent_size)

    _, raw_hash = self.deserializer.decoder.DecodeUint32Varint()
    hash_algorithm = definitions.CryptoKeyAlgorithm(raw_hash)

    algorithm_parameters = {
        'id': crypto_key_algorithm,
        'modulus_length_bits': modulus_length_bits,
        'public_exponent_size': public_exponent_size,
        'public_exponent_bytes': public_exponent_bytes,
        'hash': hash_algorithm
    }

    return key_type, algorithm_parameters

  def _ReadECKey(
      self
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads an EC Key parameters from the current decoder position.

    Returns:
      The EC Key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(raw_key_type)

    _, raw_named_curve = self.deserializer.decoder.DecodeUint32Varint()
    named_curve_type = definitions.NamedCurve(raw_named_curve)

    algorithm_parameters = {
        'crypto_key_algorithm': crypto_key_algorithm,
        'named_curve_type': named_curve_type
    }

    return key_type, algorithm_parameters

  def _ReadED25519Key(
      self
  ) -> tuple[definitions.AsymmetricCryptoKeyType, dict[str, Any]]:
    """Reads a ED25519 CryptoKey from the current decoder position.

    Returns:
      The ED25519 key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    _, raw_key_type = self.deserializer.decoder.DecodeUint32Varint()
    key_type = definitions.AsymmetricCryptoKeyType(
        raw_key_type)

    algorithm_parameters = {
        'crypto_key_algorithm': crypto_key_algorithm
    }

    return key_type, algorithm_parameters

  def ReadNoParamsKey(
      self
  ) -> tuple[definitions.WebCryptoKeyType, dict[str, Any]]:
    """Reads a No Params CryptoKey from the current decoder position.

    Returns:
      The No Parameters key algorithm parameters.
    """
    _, raw_id = self.deserializer.decoder.DecodeUint32Varint()
    crypto_key_algorithm = definitions.CryptoKeyAlgorithm(raw_id)

    algorithm_parameters = {
        'crypto_key_algorithm': crypto_key_algorithm
    }

    return definitions.WebCryptoKeyType.SECRET, algorithm_parameters

  def _ReadBlob(self) -> Optional[Blob]:
    """Reads a Blob from the current position.

    Returns:
      A Blob if the version is less then 3, None otherwise.
    """
    if self.version and self.version < 3:
      return None

    uuid = self.deserializer.ReadUTF8String()
    blob_type = self.deserializer.ReadUTF8String()
    _, size = self.deserializer.decoder.DecodeUint64Varint()
    return Blob(uuid=uuid, type=blob_type, size=size)

  def _ReadBlobIndex(self) -> Optional[BlobIndex]:
    """Reads a BlobIndex from the current decoder position.

    Returns:
      A parsed BlobIndex if the version is greater than 6, None otherwise.
    """
    if self.version < 6:
      return None
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return BlobIndex(index=index)

  def _ReadFile(self) -> Optional[File]:
    """Reads a Javascript File from the current position.

    Returns:
      A File instance, None if the version is less than 3.
    """
    if self.version < 3:
      return None

    path = self.deserializer.ReadUTF8String()
    name = self.deserializer.ReadUTF8String() if self.version >= 4 else None
    relative_path = (
        self.deserializer.ReadUTF8String() if self.version >= 4 else None)
    uuid = self.deserializer.ReadUTF8String()
    file_type = self.deserializer.ReadUTF8String()
    has_snapshot = (
        self.deserializer.decoder.DecodeUint32Varint()[1]
        if self.version >= 4 else 0)

    if has_snapshot:
      _, size = self.deserializer.decoder.DecodeUint64Varint()
      _, last_modified_ms = self.deserializer.decoder.DecodeDouble()
      if self.version < 8:
        last_modified_ms *= _MS_PER_SECOND
    else:
      size = None
      last_modified_ms = None

    is_user_visible = (
        self.deserializer.decoder.DecodeUint32Varint()[1]
        if self.version >= 7 else 1)

    return File(
        path=path,
        name=name,
        relative_path=relative_path,
        uuid=uuid,
        type=file_type,
        has_snapshot=has_snapshot,
        size=size,
        last_modified_ms=last_modified_ms,
        is_user_visible=is_user_visible)

  def _ReadFileIndex(self) -> Optional[FileIndex]:
    """Reads a FileIndex from the current position.

    Returns:
      A FileIndex, or None if the version is less than 6.
    """
    if self.version < 6:
      return None
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return FileIndex(index=index)

  def _ReadFileList(self) -> Optional[FileList]:
    """Reads a Javascript FileList from the current position.

    Returns:
      A FileList, or None if a File could not be read.
    """
    _, length = self.deserializer.decoder.DecodeUint32Varint()
    files = []
    for _ in range(length):
      decoded_file = self._ReadFile()
      if not decoded_file:
        return None
      files.append(decoded_file)
    return FileList(files=files)

  def _ReadFileListIndex(self) -> Optional[FileListIndex]:
    """Reads a Javascript FileListIndex from the current position.

    Returns:
      A FileListIndex, or None if a FileIndex could not be read.
    """
    _, length = self.deserializer.decoder.DecodeUint32Varint()
    file_indexes = []
    for _ in range(length):
      decoded_file_index = self._ReadFileIndex()
      if not decoded_file_index:
        return None
      file_indexes.append(decoded_file_index)
    return FileListIndex(file_indexes=file_indexes)

  def _ReadImageBitmap(self):
    """Reads an ImageBitmap."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadImageBitmap')

  def _ReadImageBitmapTransfer(self) -> ImageBitmapTransfer:
    """Reads an ImageBitmapTransfer."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return ImageBitmapTransfer(index=index)

  def _ReadImageData(self):
    """Reads an ImageData from the current position."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadImageData')

  def _ReadDOMPoint(self) -> DOMPoint:
    """Reads a DOMPoint from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, z = self.deserializer.decoder.DecodeDouble()
    _, w = self.deserializer.decoder.DecodeDouble()
    return DOMPoint(x=x, y=y, z=z, w=w)

  def _ReadDOMPointReadOnly(self) -> DOMPointReadOnly:
    """Reads a DOMPointReadOnly from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, z = self.deserializer.decoder.DecodeDouble()
    _, w = self.deserializer.decoder.DecodeDouble()
    return DOMPointReadOnly(x=x, y=y, z=z, w=w)

  def _ReadDOMRect(self) -> DOMRect:
    """Reads a DOMRect from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, width = self.deserializer.decoder.DecodeDouble()
    _, height = self.deserializer.decoder.DecodeDouble()
    return DOMRect(x=x, y=y, width=width, height=height)

  def _ReadDOMRectReadOnly(self) -> DOMRectReadOnly:
    """Reads a DOMRectReadOnly from the current position."""
    _, x = self.deserializer.decoder.DecodeDouble()
    _, y = self.deserializer.decoder.DecodeDouble()
    _, width = self.deserializer.decoder.DecodeDouble()
    _, height = self.deserializer.decoder.DecodeDouble()
    return DOMRectReadOnly(x=x, y=y, width=width, height=height)

  def _ReadDOMQuad(self) -> DOMQuad:
    """Reads a DOMQuad from the current position."""
    p1 = self._ReadDOMPoint()
    p2 = self._ReadDOMPoint()
    p3 = self._ReadDOMPoint()
    p4 = self._ReadDOMPoint()
    return DOMQuad(p1=p1, p2=p2, p3=p3, p4=p4)

  def _ReadDOMMatrix2D(self) -> DOMMatrix2D:
    """Reads a Javascript DOMMatrix2D from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(6)]
    return DOMMatrix2D(values=values)

  def _ReadDOMMatrix2DReadOnly(self) -> DOMMatrix2DReadOnly:
    """Reads a Javascript Read-Only DOMMatrix2D from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(6)]
    return DOMMatrix2DReadOnly(values=values)

  def _ReadDOMMatrix(self) -> DOMMatrix:
    """Reads a Javascript DOMMatrix from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(16)]
    return DOMMatrix(values=values)

  def _ReadDOMMatrixReadOnly(self) -> DOMMatrixReadOnly:
    """Reads a Javascript Read-Only DOMMatrix from the current position."""
    values = [self.deserializer.decoder.DecodeDouble()[1] for _ in range(16)]
    return DOMMatrixReadOnly(values=values)

  def _ReadMessagePort(self) -> MessagePort:
    """Reads a MessagePort from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MessagePort(index=index)

  def _ReadMojoHandle(self) -> MojoHandle:
    """Reads a MojoHandle from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MojoHandle(index=index)

  def _ReadOffscreenCanvasTransfer(self):
    """Reads a Offscreen Canvas Transfer from the current position."""
    _, width = self.deserializer.decoder.DecodeUint32Varint()
    _, height = self.deserializer.decoder.DecodeUint32Varint()
    _, canvas_id = self.deserializer.decoder.DecodeUint32Varint()
    _, client_id = self.deserializer.decoder.DecodeUint32Varint()
    _, sink_id = self.deserializer.decoder.DecodeUint32Varint()
    _, filter_quality = self.deserializer.decoder.DecodeUint32Varint()

    return OffscreenCanvasTransfer(
        width=width,
        height=height,
        canvas_id=canvas_id,
        client_id=client_id,
        sink_id=sink_id,
        filter_quality=filter_quality)

  def _ReadReadableStreamTransfer(self) -> ReadableStreamTransfer:
    """Reads a Readable Stream Transfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return ReadableStreamTransfer(index=index)

  def _ReadWriteableStreamTransfer(self) -> WriteableStreamTransfer:
    """Reads a Writeable Stream Transfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return WriteableStreamTransfer(index=index)

  def _ReadTransformStreamTransfer(self) -> TransformStreamTransfer:
    """Reads a TransformStreamTransfer from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return TransformStreamTransfer(index=index)

  def _ReadDOMException(self) -> DOMException:
    """Reads a DOMException from the current position."""
    name = self.deserializer.ReadUTF8String()
    message = self.deserializer.ReadUTF8String()
    stack_unused = self.deserializer.ReadUTF8String()
    return DOMException(name=name, message=message, stack_unused=stack_unused)

  def _ReadRTCEncodedAudioFrame(self) -> RTCEncodedAudioFrame:
    """Reads a RTC Encoded Audio Frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return RTCEncodedAudioFrame(index=index)

  def _ReadRTCEncodedVideoFrame(self) -> RTCEncodedVideoFrame:
    """Reads a RTC Encoded Video Frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return RTCEncodedVideoFrame(index=index)

  def _ReadCryptoKey(self) -> CryptoKey:
    """Reads a CryptoKey from the current position.

    Returns:
      A parsed CryptoKey.
    """
    _, raw_key_byte = self.deserializer.decoder.DecodeUint8()
    key_byte = definitions.CryptoKeySubTag(raw_key_byte)
    if key_byte == definitions.CryptoKeySubTag.AES_KEY:
      key_type, algorithm_parameters = self._ReadAESKey()
    elif key_byte == definitions.CryptoKeySubTag.HMAC_KEY:
      key_type, algorithm_parameters = self._ReadHMACKey()
    elif key_byte == definitions.CryptoKeySubTag.RSA_HASHED_KEY:
      key_type, algorithm_parameters = self._ReadRSAHashedKey()
    elif key_byte == definitions.CryptoKeySubTag.EC_KEY:
      key_type, algorithm_parameters = self._ReadECKey()
    elif key_byte == definitions.CryptoKeySubTag.ED25519_KEY:
      key_type, algorithm_parameters = self._ReadED25519Key()
    elif key_byte == definitions.CryptoKeySubTag.NO_PARAMS_KEY:
      key_type, algorithm_parameters = self.ReadNoParamsKey()

    _, raw_usages = self.deserializer.decoder.DecodeUint32Varint()
    usages = definitions.CryptoKeyUsage(raw_usages)

    extractable = bool(raw_usages & definitions.CryptoKeyUsage.EXTRACTABLE)
    _, key_data_length = self.deserializer.decoder.DecodeUint32Varint()

    _, key_data = self.deserializer.decoder.ReadBytes(count=key_data_length)

    return CryptoKey(
        key_type=key_type,
        algorithm_parameters=algorithm_parameters,
        extractable=extractable,
        usages=usages,
        key_data=key_data
    )

  def _ReadAudioData(self) -> AudioData:
    """Reads an AudioData from the current position."""
    _, audio_frame_index = self.deserializer.decoder.DecodeUint32Varint()
    return AudioData(audio_frame_index=audio_frame_index)

  def _ReadDomFileSystem(self) -> DOMFileSystem:
    """Reads an DOMFileSystem from the current position."""
    _, raw_type = self.deserializer.decoder.DecodeUint32Varint()
    name = self.deserializer.ReadUTF8String()
    root_url = self.deserializer.ReadUTF8String()
    return DOMFileSystem(raw_type=raw_type, name=name, root_url=root_url)

  def _ReadFileSystemFileHandle(self) -> FileSystemHandle:
    """Reads a FileSystemHandle from the current position."""
    name = self.deserializer.ReadUTF8String()
    _, token_index = self.deserializer.decoder.DecodeUint32Varint()
    return FileSystemHandle(name=name, token_index=token_index)

  def _ReadVideoFrame(self) -> VideoFrame:
    """Reads the video frame from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return VideoFrame(index=index)

  def _ReadEncodedAudioChunk(self) -> EncodedAudioChunk:
    """Reads the encoded audio chunk from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return EncodedAudioChunk(index=index)

  def _ReadEncodedVideoChunk(self) -> EncodedVideoChunk:
    """Reads the encoded video chunk from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return EncodedVideoChunk(index=index)

  def _ReadMediaStreamTrack(self):
    """Reads the media stream track from the current position."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadMediaStreamTrack')

  def _ReadCropTarget(self):
    """Reads the crop target from the current position."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadCropTarget')

  def _ReadRestrictionTarget(self):
    """Reads the restriction target from the current position."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadRestrictionTarget')

  def _ReadMediaSourceHandle(self) -> MediaSourceHandle:
    """Reads the media source handle from the current position."""
    _, index = self.deserializer.decoder.DecodeUint32Varint()
    return MediaSourceHandle(index=index)

  def _ReadFencedFrameConfig(self):
    """Reads the fenced frame target from the current position."""
    raise NotImplementedError('V8ScriptValueDecoder._ReadFencedFrameConfig')

  def ReadTag(self) -> definitions.BlinkSerializationTag:
    """Reads a blink serialization tag from the current position.

    Returns:
      The blink serialization tag.

    Raises:
      ParserError: if an invalid blink tag is read.
    """
    _, tag_value = self.deserializer.decoder.DecodeUint8()
    try:
      return definitions.BlinkSerializationTag(tag_value)
    except ValueError as err:
      offset = self.deserializer.decoder.stream.tell()
      raise errors.ParserError(
          f'Invalid blink tag encountered at offset {offset}') from err

  def ReadHostObject(self) -> Any:
    """Reads a host object from the current position.

    Returns:
      The Host Object.
    """
    tag = self.ReadTag()
    dom_object = None

    # V8ScriptValueDeserializer
    if tag == definitions.BlinkSerializationTag.BLOB:
      dom_object = self._ReadBlob()
    elif tag == definitions.BlinkSerializationTag.BLOB_INDEX:
      dom_object = self._ReadBlobIndex()
    elif tag == definitions.BlinkSerializationTag.FILE:
      dom_object = self._ReadFile()
    elif tag == definitions.BlinkSerializationTag.FILE_INDEX:
      dom_object = self._ReadFileIndex()
    elif tag == definitions.BlinkSerializationTag.FILE_LIST:
      dom_object = self._ReadFileList()
    elif tag == definitions.BlinkSerializationTag.FILE_LIST_INDEX:
      dom_object = self._ReadFileListIndex()
    elif tag == definitions.BlinkSerializationTag.IMAGE_BITMAP:
      dom_object = self._ReadImageBitmap()
    elif tag == definitions.BlinkSerializationTag.IMAGE_BITMAP_TRANSFER:
      dom_object = self._ReadImageBitmapTransfer()
    elif tag == definitions.BlinkSerializationTag.IMAGE_DATA:
      dom_object = self._ReadImageData()
    elif tag == definitions.BlinkSerializationTag.DOM_POINT:
      dom_object = self._ReadDOMPoint()
    elif tag ==  definitions.BlinkSerializationTag.DOM_POINT_READ_ONLY:
      dom_object = self._ReadDOMPointReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_RECT:
      dom_object = self._ReadDOMRect()
    elif tag == definitions.BlinkSerializationTag.DOM_RECT_READ_ONLY:
      dom_object = self._ReadDOMRectReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_QUAD:
      dom_object = self._ReadDOMQuad()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX2D:
      dom_object = self._ReadDOMMatrix2D()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX2D_READ_ONLY:
      dom_object = self._ReadDOMMatrix2DReadOnly()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX:
      dom_object = self._ReadDOMMatrix()
    elif tag == definitions.BlinkSerializationTag.DOM_MATRIX_READ_ONLY:
      dom_object = self._ReadDOMMatrixReadOnly()
    elif tag == definitions.BlinkSerializationTag.MESSAGE_PORT:
      dom_object = self._ReadMessagePort()
    elif tag == definitions.BlinkSerializationTag.MOJO_HANDLE:
      dom_object = self._ReadMojoHandle()
    elif tag == definitions.BlinkSerializationTag.OFFSCREEN_CANVAS_TRANSFER:
      dom_object = self._ReadOffscreenCanvasTransfer()
    elif tag == definitions.BlinkSerializationTag.READABLE_STREAM_TRANSFER:
      dom_object = self._ReadReadableStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.WRITABLE_STREAM_TRANSFER:
      dom_object = self._ReadWriteableStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.TRANSFORM_STREAM_TRANSFER:
      dom_object = self._ReadTransformStreamTransfer()
    elif tag == definitions.BlinkSerializationTag.DOM_EXCEPTION:
      dom_object = self._ReadDOMException()

    # V8ScriptValueDeserializerForModules
    elif tag == definitions.BlinkSerializationTag.CRYPTO_KEY:
      dom_object = self._ReadCryptoKey()
    elif tag == definitions.BlinkSerializationTag.DOM_FILE_SYSTEM:
      dom_object = self._ReadDomFileSystem()
    elif tag == definitions.BlinkSerializationTag.FILE_SYSTEM_FILE_HANDLE:
      dom_object = self._ReadFileSystemFileHandle()
    elif tag == definitions.BlinkSerializationTag.RTC_ENCODED_AUDIO_FRAME:
      dom_object = self._ReadRTCEncodedAudioFrame()
    elif tag == definitions.BlinkSerializationTag.RTC_ENCODED_VIDEO_FRAME:
      dom_object = self._ReadRTCEncodedVideoFrame()
    elif tag == definitions.BlinkSerializationTag.AUDIO_DATA:
      dom_object = self._ReadAudioData()
    elif tag == definitions.BlinkSerializationTag.VIDEO_FRAME:
      dom_object = self._ReadVideoFrame()
    elif tag == definitions.BlinkSerializationTag.ENCODED_AUDIO_CHUNK:
      dom_object = self._ReadEncodedAudioChunk()
    elif tag == definitions.BlinkSerializationTag.ENCODED_VIDEO_CHUNK:
      dom_object = self._ReadEncodedVideoChunk()
    elif tag == definitions.BlinkSerializationTag.MEDIA_STREAM_TRACK:
      dom_object = self._ReadMediaStreamTrack()
    elif tag == definitions.BlinkSerializationTag.CROP_TARGET:
      dom_object = self._ReadCropTarget()
    elif tag == definitions.BlinkSerializationTag.RESTRICTION_TARGET:
      dom_object = self._ReadRestrictionTarget()
    elif tag == definitions.BlinkSerializationTag.MEDIA_SOURCE_HANDLE:
      dom_object = self._ReadMediaSourceHandle()
    elif tag == definitions.BlinkSerializationTag.FENCED_FRAME_CONFIG:
      dom_object = self._ReadFencedFrameConfig()
    return dom_object

  def Deserialize(self) -> Any:
    """Deserializes a Blink SSV.

    The serialization format has two 'envelopes'.
    [version tag] [Blink version] [version tag] [v8 version] ...

    Returns:
      The deserialized script value.

    Raises:
      ParserError: if an unsupported header version was found.
    """
    version_envelope_size = self._ReadVersionEnvelope()
    if self.trailer_size:
      self.deserializer = v8.ValueDeserializer(
          io.BytesIO(
              self.raw_data[version_envelope_size:self.trailer_offset]),
          delegate=self)
    else:
      self.deserializer = v8.ValueDeserializer(
          io.BytesIO(
              self.raw_data[version_envelope_size:]),
          delegate=self)
    is_supported = self.deserializer.ReadHeader()
    if not is_supported:
      raise errors.ParserError('Unsupported header')
    return self.deserializer.ReadValue()

  @classmethod
  def FromBytes(cls, data: bytes) -> Any:
    """Deserializes a Blink SSV from the data array.

    Returns:
      The deserialized script value.
    """
    return cls(data).Deserialize()
